from copy import deepcopy
from typing import NamedTuple

import numpy as np
import pymc as pm
import pytensor.tensor as pt
import pytest
import xarray as xr

from graphviz.graphs import Digraph
from preliz.distributions import distributions as preliz_distributions
from pydantic import ValidationError
from pymc.model_graph import fast_eval

import pymc_extras.prior as pr

from pymc_extras.deserialize import (
    DESERIALIZERS,
    deserialize,
    register_deserialization,
)
from pymc_extras.prior import (
    Censored,
    MuAlreadyExistsError,
    Prior,
    Scaled,
    UnknownTransformError,
    UnsupportedDistributionError,
    UnsupportedParameterizationError,
    UnsupportedShapeError,
    VariableFactory,
    handle_dims,
    register_tensor_transform,
    sample_prior,
)


@pytest.mark.parametrize(
    "x, dims, desired_dims, expected_fn",
    [
        (np.arange(3), "channel", "channel", lambda x: x),
        (np.arange(3), "channel", ("geo", "channel"), lambda x: x),
        (np.arange(3), "channel", ("channel", "geo"), lambda x: x[:, None]),
        (np.arange(3), "channel", ("x", "y", "channel", "geo"), lambda x: x[:, None]),
        (
            np.arange(3 * 2).reshape(3, 2),
            ("channel", "geo"),
            ("geo", "x", "y", "channel"),
            lambda x: x.T[:, None, None, :],
        ),
        (
            np.arange(4 * 2 * 3).reshape(4, 2, 3),
            ("channel", "geo", "store"),
            ("geo", "x", "store", "channel"),
            lambda x: x.swapaxes(0, 2).swapaxes(0, 1)[:, None, :, :],
        ),
    ],
    ids=[
        "same_dims",
        "different_dims",
        "dim_padding",
        "just_enough_dims",
        "transpose_and_padding",
        "swaps_and_padding",
    ],
)
def test_handle_dims(x, dims, desired_dims, expected_fn) -> None:
    result = handle_dims(x, dims, desired_dims)
    if isinstance(result, pt.TensorVariable):
        result = fast_eval(result)

    np.testing.assert_array_equal(result, expected_fn(x))


@pytest.mark.parametrize(
    "x, dims, desired_dims",
    [
        (np.ones(3), "channel", "something_else"),
        (np.ones((3, 2)), ("a", "b"), ("a", "B")),
    ],
    ids=["no_incommon", "some_incommon"],
)
def test_handle_dims_with_impossible_dims(x, dims, desired_dims) -> None:
    match = " are not a subset of the desired dims "
    with pytest.raises(UnsupportedShapeError, match=match):
        handle_dims(x, dims, desired_dims)


def test_missing_transform() -> None:
    match = "Neither pytensor.tensor nor pymc.math have the function 'foo_bar'"
    with pytest.raises(UnknownTransformError, match=match):
        Prior("Normal", transform="foo_bar")


def test_get_item() -> None:
    var = Prior("Normal", mu=0, sigma=1)

    assert var["mu"] == 0
    assert var["sigma"] == 1


def test_noncentered_needs_params() -> None:
    with pytest.raises(ValueError):
        Prior(
            "Normal",
            centered=False,
        )


def test_different_than_pymc_params() -> None:
    with pytest.raises(ValueError):
        Prior("Normal", mu=0, b=1)


def test_non_unique_dims() -> None:
    with pytest.raises(ValueError):
        Prior("Normal", mu=0, sigma=1, dims=("channel", "channel"))


def test_doesnt_check_validity_parameterization() -> None:
    try:
        Prior("Normal", mu=0, sigma=1, tau=1)
    except Exception as e:
        pytest.fail(f"Unexpected exception: {e}")


def test_doesnt_check_validity_values() -> None:
    try:
        Prior("Normal", mu=0, sigma=-1)
    except Exception as e:
        pytest.fail(f"Unexpected exception: {e}")


def test_preliz() -> None:
    var = Prior("Normal", mu=0, sigma=1)
    dist = var.preliz
    assert isinstance(dist, preliz_distributions.Distribution)


@pytest.mark.parametrize(
    "var, expected",
    [
        (Prior("Normal", mu=0, sigma=1), 'Prior("Normal", mu=0, sigma=1)'),
        (
            Prior("Normal", mu=Prior("Normal"), sigma=Prior("HalfNormal")),
            'Prior("Normal", mu=Prior("Normal"), sigma=Prior("HalfNormal"))',
        ),
        (Prior("Normal", dims="channel"), 'Prior("Normal", dims="channel")'),
        (
            Prior("Normal", mu=0, sigma=1, transform="sigmoid"),
            'Prior("Normal", mu=0, sigma=1, transform="sigmoid")',
        ),
    ],
)
def test_str(var, expected) -> None:
    assert str(var) == expected


@pytest.mark.parametrize(
    "var",
    [
        Prior("Normal", mu=0, sigma=1),
        Prior("Normal", mu=Prior("Normal"), sigma=Prior("HalfNormal"), dims="channel"),
        Prior("Normal", dims=("geo", "channel")),
    ],
)
def test_repr(var) -> None:
    assert eval(repr(var)) == var


def test_invalid_distribution() -> None:
    with pytest.raises(UnsupportedDistributionError):
        Prior("Invalid")


def test_broadcast_doesnt_work():
    with pytest.raises(UnsupportedShapeError):
        Prior(
            "Normal",
            mu=0,
            sigma=Prior("HalfNormal", sigma=1, dims="x"),
            dims="y",
        )


def test_dim_workaround_flaw() -> None:
    distribution = Prior(
        "Normal",
        mu=Prior("Normal"),
        sigma=Prior("HalfNormal"),
        dims="y",
    )

    try:
        distribution["mu"].dims = "x"
    except Exception as e:
        pytest.fail(f"Unexpected exception: {e}")

    with pytest.raises(UnsupportedShapeError):
        distribution._param_dims_work()


def test_noncentered_error() -> None:
    with pytest.raises(UnsupportedParameterizationError):
        Prior(
            "Gamma",
            mu=0,
            sigma=1,
            dims="x",
            centered=False,
        )


def test_create_variable_multiple_times() -> None:
    mu = Prior(
        "Normal",
        mu=Prior("Normal"),
        sigma=Prior("HalfNormal"),
        dims="channel",
        centered=False,
    )

    coords = {
        "channel": ["a", "b", "c"],
    }
    with pm.Model(coords=coords) as model:
        mu.create_variable("mu")
        mu.create_variable("mu_2")

    suffixes = [
        "",
        "_offset",
        "_mu",
        "_sigma",
    ]
    dims = [(3,), (3,), (), ()]

    for prefix in ["mu", "mu_2"]:
        for suffix, dim in zip(suffixes, dims, strict=False):
            assert fast_eval(model[f"{prefix}{suffix}"]).shape == dim


@pytest.fixture
def large_var() -> Prior:
    mu = Prior(
        "Normal",
        mu=Prior("Normal", mu=1),
        sigma=Prior("HalfNormal"),
        dims="channel",
        centered=False,
    )
    sigma = Prior("HalfNormal", sigma=Prior("HalfNormal"), dims="geo")

    return Prior("Normal", mu=mu, sigma=sigma, dims=("geo", "channel"))


def test_create_variable(large_var) -> None:
    coords = {
        "channel": ["a", "b", "c"],
        "geo": ["x", "y"],
    }
    with pm.Model(coords=coords) as model:
        large_var.create_variable("var")

    var_names = [
        "var",
        "var_mu",
        "var_sigma",
        "var_mu_offset",
        "var_mu_mu",
        "var_mu_sigma",
        "var_sigma_sigma",
    ]
    assert set(var.name for var in model.unobserved_RVs) == set(var_names)
    dims = [
        (2, 3),
        (3,),
        (2,),
        (3,),
        (),
        (),
        (),
    ]
    for var_name, dim in zip(var_names, dims, strict=False):
        assert fast_eval(model[var_name]).shape == dim


def test_transform() -> None:
    var = Prior("Normal", mu=0, sigma=1, transform="sigmoid")

    with pm.Model() as model:
        var.create_variable("var")

    var_names = [
        "var",
        "var_raw",
    ]
    dims = [
        (),
        (),
    ]
    for var_name, dim in zip(var_names, dims, strict=False):
        assert fast_eval(model[var_name]).shape == dim


def test_to_dict(large_var) -> None:
    data = large_var.to_dict()

    assert data == {
        "dist": "Normal",
        "kwargs": {
            "mu": {
                "dist": "Normal",
                "kwargs": {
                    "mu": {
                        "dist": "Normal",
                        "kwargs": {
                            "mu": 1,
                        },
                    },
                    "sigma": {
                        "dist": "HalfNormal",
                    },
                },
                "centered": False,
                "dims": ("channel",),
            },
            "sigma": {
                "dist": "HalfNormal",
                "kwargs": {
                    "sigma": {
                        "dist": "HalfNormal",
                    },
                },
                "dims": ("geo",),
            },
        },
        "dims": ("geo", "channel"),
    }


def test_to_dict_numpy() -> None:
    var = Prior("Normal", mu=np.array([0, 10, 20]), dims="channel")
    assert var.to_dict() == {
        "dist": "Normal",
        "kwargs": {
            "mu": [0, 10, 20],
        },
        "dims": ("channel",),
    }


def test_dict_round_trip(large_var) -> None:
    assert Prior.from_dict(large_var.to_dict()) == large_var


def test_constrain_with_transform_error() -> None:
    var = Prior("Normal", transform="sigmoid")

    with pytest.raises(ValueError):
        var.constrain(lower=0, upper=1)


def test_constrain(mocker) -> None:
    var = Prior("Normal")

    mocker.patch(
        "preliz.maxent",
        return_value=mocker.Mock(params_dict={"mu": 5, "sigma": 2}),
    )

    new_var = var.constrain(lower=0, upper=1)
    assert new_var == Prior("Normal", mu=5, sigma=2)


def test_dims_change() -> None:
    var = Prior("Normal", mu=0, sigma=1)
    var.dims = "channel"

    assert var.dims == ("channel",)


def test_dims_change_error() -> None:
    mu = Prior("Normal", dims="channel")
    var = Prior("Normal", mu=mu, dims="channel")

    with pytest.raises(UnsupportedShapeError):
        var.dims = "geo"


def test_deepcopy() -> None:
    priors = {
        "alpha": Prior("Beta", alpha=1, beta=1),
        "gamma": Prior("Normal", mu=0, sigma=1),
    }

    new_priors = deepcopy(priors)
    priors["alpha"].dims = "channel"

    assert new_priors["alpha"].dims == ()


@pytest.fixture
def mmm_default_model_config():
    return {
        "intercept": {"dist": "Normal", "kwargs": {"mu": 0, "sigma": 2}},
        "likelihood": {
            "dist": "Normal",
            "kwargs": {
                "sigma": {"dist": "HalfNormal", "kwargs": {"sigma": 2}},
            },
        },
        "gamma_control": {
            "dist": "Normal",
            "kwargs": {"mu": 0, "sigma": 2},
            "dims": "control",
        },
        "gamma_fourier": {
            "dist": "Laplace",
            "kwargs": {"mu": 0, "b": 1},
            "dims": "fourier_mode",
        },
    }


def test_backwards_compat(mmm_default_model_config) -> None:
    result = {param: Prior.from_dict(value) for param, value in mmm_default_model_config.items()}
    assert result == {
        "intercept": Prior("Normal", mu=0, sigma=2),
        "likelihood": Prior("Normal", sigma=Prior("HalfNormal", sigma=2)),
        "gamma_control": Prior("Normal", mu=0, sigma=2, dims="control"),
        "gamma_fourier": Prior("Laplace", mu=0, b=1, dims="fourier_mode"),
    }


def test_sample_prior() -> None:
    var = Prior(
        "Normal",
        mu=Prior("Normal"),
        sigma=Prior("HalfNormal"),
        dims="channel",
        transform="sigmoid",
    )

    coords = {"channel": ["A", "B", "C"]}
    prior = var.sample_prior(coords=coords, draws=25)

    assert isinstance(prior, xr.Dataset)
    assert prior.sizes == {"chain": 1, "draw": 25, "channel": 3}


def test_sample_prior_missing_coords() -> None:
    dist = Prior("Normal", dims="channel")

    with pytest.raises(KeyError, match="Coords"):
        dist.sample_prior()


def test_to_graph() -> None:
    hierarchical_distribution = Prior(
        "Normal",
        mu=Prior("Normal"),
        sigma=Prior("HalfNormal"),
        dims="channel",
    )

    G = hierarchical_distribution.to_graph()
    assert isinstance(G, Digraph)


def test_from_dict_list() -> None:
    data = {
        "dist": "Normal",
        "kwargs": {
            "mu": [0, 1, 2],
            "sigma": 1,
        },
        "dims": "channel",
    }

    var = Prior.from_dict(data)
    assert var.dims == ("channel",)
    assert isinstance(var["mu"], np.ndarray)
    np.testing.assert_array_equal(var["mu"], [0, 1, 2])


def test_from_dict_list_dims() -> None:
    data = {
        "dist": "Normal",
        "kwargs": {
            "mu": 0,
            "sigma": 1,
        },
        "dims": ["channel", "geo"],
    }

    var = Prior.from_dict(data)
    assert var.dims == ("channel", "geo")


def test_to_dict_transform() -> None:
    dist = Prior("Normal", transform="sigmoid")

    data = dist.to_dict()
    assert data == {
        "dist": "Normal",
        "transform": "sigmoid",
    }


def test_equality_non_prior() -> None:
    dist = Prior("Normal")

    assert dist != 1


def test_deepcopy_memo() -> None:
    memo = {}
    dist = Prior("Normal")
    memo[id(dist)] = dist
    deepcopy(dist, memo)
    assert len(memo) == 1
    deepcopy(dist, memo)

    assert len(memo) == 1


def test_create_likelihood_variable() -> None:
    distribution = Prior("Normal", sigma=Prior("HalfNormal"))

    with pm.Model() as model:
        mu = pm.Normal("mu")

        data = distribution.create_likelihood_variable("data", mu=mu, observed=10)

    assert model.observed_RVs == [data]
    assert "data_sigma" in model


def test_create_likelihood_variable_already_has_mu() -> None:
    distribution = Prior("Normal", mu=Prior("Normal"), sigma=Prior("HalfNormal"))

    with pm.Model():
        mu = pm.Normal("mu")

        with pytest.raises(MuAlreadyExistsError):
            distribution.create_likelihood_variable("data", mu=mu, observed=10)


def test_create_likelihood_non_mu_parameterized_distribution() -> None:
    distribution = Prior("Cauchy")

    with pm.Model():
        mu = pm.Normal("mu")
        with pytest.raises(UnsupportedDistributionError):
            distribution.create_likelihood_variable("data", mu=mu, observed=10)


def test_non_centered_student_t() -> None:
    try:
        Prior(
            "StudentT",
            mu=Prior("Normal"),
            sigma=Prior("HalfNormal"),
            nu=Prior("HalfNormal"),
            dims="channel",
            centered=False,
        )
    except Exception as e:
        pytest.fail(f"Unexpected exception: {e}")


def test_cant_reset_distribution() -> None:
    dist = Prior("Normal")
    with pytest.raises(AttributeError, match="Can't change the distribution"):
        dist.distribution = "Cauchy"


def test_nonstring_distribution() -> None:
    with pytest.raises(ValidationError, match=".*Input should be a valid string.*"):
        Prior(pm.Normal)


def test_change_the_transform() -> None:
    dist = Prior("Normal")
    dist.transform = "logit"
    assert dist.transform == "logit"


def test_nonstring_transform() -> None:
    with pytest.raises(ValidationError, match=".*Input should be a valid string.*"):
        Prior("Normal", transform=pm.math.log)


def test_checks_param_value_types() -> None:
    with pytest.raises(ValueError, match="Parameters must be one of the following types"):
        Prior("Normal", mu="str", sigma="str")


def test_check_equality_with_numpy() -> None:
    dist = Prior("Normal", mu=np.array([1, 2, 3]), sigma=1)
    assert dist == dist.deepcopy()


def clear_custom_transforms() -> None:
    global CUSTOM_TRANSFORMS
    CUSTOM_TRANSFORMS = {}


def test_custom_transform() -> None:
    new_transform_name = "foo_bar"
    with pytest.raises(UnknownTransformError):
        Prior("Normal", transform=new_transform_name)

    register_tensor_transform(new_transform_name, lambda x: x**2)

    dist = Prior("Normal", transform=new_transform_name)
    prior = dist.sample_prior(draws=10)
    df_prior = prior.to_dataframe()

    np.testing.assert_array_equal(
        df_prior.variable.to_numpy(),
        df_prior.variable_raw.to_numpy() ** 2,
    )


def test_custom_transform_comes_first() -> None:
    # function in pytensor.tensor
    register_tensor_transform("square", lambda x: 2 * x)

    dist = Prior("Normal", transform="square")
    prior = dist.sample_prior(draws=10)
    df_prior = prior.to_dataframe()

    np.testing.assert_array_equal(
        df_prior.variable.to_numpy(),
        2 * df_prior.variable_raw.to_numpy(),
    )

    clear_custom_transforms()


def test_serialize_with_pytensor() -> None:
    sigma = pt.arange(1, 4)
    dist = Prior("Normal", mu=0, sigma=sigma)

    assert dist.to_dict() == {
        "dist": "Normal",
        "kwargs": {
            "mu": 0,
            "sigma": [1, 2, 3],
        },
    }


def test_zsn_non_centered() -> None:
    try:
        Prior("ZeroSumNormal", sigma=1, centered=False)
    except Exception as e:
        pytest.fail(f"Unexpected exception: {e}")


class Arbitrary:
    def __init__(self, dims: str | tuple[str, ...]) -> None:
        self.dims = dims

    def create_variable(self, name: str):
        return pm.Normal(name, dims=self.dims)


class ArbitraryWithoutName:
    def __init__(self, dims: str | tuple[str, ...]) -> None:
        self.dims = dims

    def create_variable(self, name: str):
        with pm.Model(name=name):
            location = pm.Normal("location", dims=self.dims)
            scale = pm.HalfNormal("scale", dims=self.dims)

            return pm.Normal("standard_normal") * scale + location


def test_sample_prior_arbitrary() -> None:
    var = Arbitrary(dims="channel")

    prior = sample_prior(var, coords={"channel": ["A", "B", "C"]}, draws=25)

    assert isinstance(prior, xr.Dataset)


def test_sample_prior_arbitrary_no_name() -> None:
    var = ArbitraryWithoutName(dims="channel")

    prior = sample_prior(var, coords={"channel": ["A", "B", "C"]}, draws=25)

    assert isinstance(prior, xr.Dataset)
    assert "variable" not in prior

    prior_with = sample_prior(
        var,
        coords={"channel": ["A", "B", "C"]},
        draws=25,
        wrap=True,
    )

    assert isinstance(prior_with, xr.Dataset)
    assert "variable" in prior_with


def test_create_prior_with_arbitrary() -> None:
    dist = Prior(
        "Normal",
        mu=Arbitrary(dims=("channel",)),
        sigma=1,
        dims=("channel", "geo"),
    )

    coords = {
        "channel": ["C1", "C2", "C3"],
        "geo": ["G1", "G2"],
    }
    with pm.Model(coords=coords) as model:
        dist.create_variable("var")

    assert "var_mu" in model
    var_mu = model["var_mu"]

    assert fast_eval(var_mu).shape == (len(coords["channel"]),)


def test_censored_is_variable_factory() -> None:
    normal = Prior("Normal")
    censored_normal = Censored(normal, lower=0)

    assert isinstance(censored_normal, VariableFactory)


@pytest.mark.parametrize(
    "dims, expected_dims",
    [
        ("channel", ("channel",)),
        (("channel", "geo"), ("channel", "geo")),
    ],
    ids=["string", "tuple"],
)
def test_censored_dims_from_distribution(dims, expected_dims) -> None:
    normal = Prior("Normal", dims=dims)
    censored_normal = Censored(normal, lower=0)

    assert censored_normal.dims == expected_dims


def test_censored_variables_created() -> None:
    normal = Prior("Normal", mu=Prior("Normal"), dims="dim")
    censored_normal = Censored(normal, lower=0)

    coords = {"dim": range(3)}
    with pm.Model(coords=coords) as model:
        censored_normal.create_variable("var")

    var_names = ["var", "var_mu"]
    assert set(var.name for var in model.unobserved_RVs) == set(var_names)
    dims = [(3,), ()]
    for var_name, dim in zip(var_names, dims, strict=False):
        assert fast_eval(model[var_name]).shape == dim


def test_censored_sample_prior() -> None:
    normal = Prior("Normal", dims="channel")
    censored_normal = Censored(normal, lower=0)

    coords = {"channel": ["A", "B", "C"]}
    prior = censored_normal.sample_prior(coords=coords, draws=25)

    assert isinstance(prior, xr.Dataset)
    assert prior.sizes == {"chain": 1, "draw": 25, "channel": 3}


def test_censored_to_graph() -> None:
    normal = Prior("Normal", dims="channel")
    censored_normal = Censored(normal, lower=0)

    G = censored_normal.to_graph()
    assert isinstance(G, Digraph)


def test_censored_likelihood_variable() -> None:
    normal = Prior("Normal", sigma=Prior("HalfNormal"), dims="channel")
    censored_normal = Censored(normal, lower=0)

    coords = {"channel": range(3)}
    with pm.Model(coords=coords) as model:
        mu = pm.Normal("mu")
        variable = censored_normal.create_likelihood_variable(
            name="likelihood",
            mu=mu,
            observed=[1, 2, 3],
        )

    assert isinstance(variable, pt.TensorVariable)
    assert model.observed_RVs == [variable]
    assert "likelihood_sigma" in model


def test_censored_likelihood_unsupported_distribution() -> None:
    cauchy = Prior("Cauchy")
    censored_cauchy = Censored(cauchy, lower=0)

    with pm.Model():
        mu = pm.Normal("mu")
        with pytest.raises(UnsupportedDistributionError):
            censored_cauchy.create_likelihood_variable(
                name="likelihood",
                mu=mu,
                observed=1,
            )


def test_censored_likelihood_already_has_mu() -> None:
    normal = Prior("Normal", mu=Prior("Normal"), sigma=Prior("HalfNormal"))
    censored_normal = Censored(normal, lower=0)

    with pm.Model():
        mu = pm.Normal("mu")
        with pytest.raises(MuAlreadyExistsError):
            censored_normal.create_likelihood_variable(
                name="likelihood",
                mu=mu,
                observed=1,
            )


def test_censored_to_dict() -> None:
    normal = Prior("Normal", mu=0, sigma=1, dims="channel")
    censored_normal = Censored(normal, lower=0)

    data = censored_normal.to_dict()
    assert data == {
        "class": "Censored",
        "data": {"dist": normal.to_dict(), "lower": 0, "upper": float("inf")},
    }


def test_deserialize_censored() -> None:
    data = {
        "class": "Censored",
        "data": {
            "dist": {
                "dist": "Normal",
            },
            "lower": 0,
            "upper": float("inf"),
        },
    }

    instance = deserialize(data)
    assert isinstance(instance, Censored)
    assert isinstance(instance.distribution, Prior)
    assert instance.lower == 0
    assert instance.upper == float("inf")


class ArbitrarySerializable(Arbitrary):
    def to_dict(self):
        return {"dims": self.dims}


@pytest.fixture
def arbitrary_serialized_data() -> dict:
    return {"dims": ("channel",)}


def test_create_prior_with_arbitrary_serializable(arbitrary_serialized_data) -> None:
    dist = Prior(
        "Normal",
        mu=ArbitrarySerializable(dims=("channel",)),
        sigma=1,
        dims=("channel", "geo"),
    )

    assert dist.to_dict() == {
        "dist": "Normal",
        "kwargs": {
            "mu": arbitrary_serialized_data,
            "sigma": 1,
        },
        "dims": ("channel", "geo"),
    }


@pytest.fixture
def register_arbitrary_deserialization():
    register_deserialization(
        lambda data: isinstance(data, dict) and data.keys() == {"dims"},
        lambda data: ArbitrarySerializable(**data),
    )

    yield

    DESERIALIZERS.pop()


def test_deserialize_arbitrary_within_prior(
    arbitrary_serialized_data,
    register_arbitrary_deserialization,
) -> None:
    data = {
        "dist": "Normal",
        "kwargs": {
            "mu": arbitrary_serialized_data,
            "sigma": 1,
        },
        "dims": ("channel", "geo"),
    }

    dist = deserialize(data)
    assert isinstance(dist["mu"], ArbitrarySerializable)
    assert dist["mu"].dims == ("channel",)


def test_censored_with_tensor_variable() -> None:
    normal = Prior("Normal", dims="channel")
    lower = pt.as_tensor_variable([0, 1, 2])
    censored_normal = Censored(normal, lower=lower)

    assert censored_normal.to_dict() == {
        "class": "Censored",
        "data": {
            "dist": normal.to_dict(),
            "lower": [0, 1, 2],
            "upper": float("inf"),
        },
    }


def test_censored_dims_setter() -> None:
    normal = Prior("Normal", dims="channel")
    censored_normal = Censored(normal, lower=0)
    censored_normal.dims = "date"
    assert normal.dims == ("date",)


class ModelData(NamedTuple):
    mu: float
    observed: list[float]


@pytest.fixture(scope="session")
def model_data() -> ModelData:
    return ModelData(mu=0, observed=[0, 1, 2, 3, 4])


@pytest.fixture(scope="session")
def normal_model_with_censored_API(model_data) -> pm.Model:
    coords = {"idx": range(len(model_data.observed))}
    with pm.Model(coords=coords) as model:
        sigma = Prior("HalfNormal")
        normal = Prior("Normal", sigma=sigma, dims="idx")
        Censored(normal, lower=0).create_likelihood_variable(
            "censored_normal",
            mu=model_data.mu,
            observed=model_data.observed,
        )

    return model


@pytest.fixture(scope="session")
def normal_model_with_censored_logp(normal_model_with_censored_API):
    return normal_model_with_censored_API.compile_logp()


@pytest.fixture(scope="session")
def expected_normal_model(model_data) -> pm.Model:
    n_points = len(model_data.observed)
    with pm.Model() as expected_model:
        sigma = pm.HalfNormal("censored_normal_sigma")
        normal = pm.Normal.dist(mu=model_data.mu, sigma=sigma, shape=n_points)
        pm.Censored(
            "censored_normal",
            normal,
            lower=0,
            upper=np.inf,
            observed=model_data.observed,
        )

    return expected_model


@pytest.fixture(scope="session")
def expected_normal_model_logp(expected_normal_model):
    return expected_normal_model.compile_logp()


@pytest.mark.parametrize("sigma_log__", [-10, -5, -2.5, 0, 2.5, 5, 10])
def test_censored_normal_logp(
    sigma_log__,
    normal_model_with_censored_logp,
    expected_normal_model_logp,
) -> None:
    points = {"censored_normal_sigma_log__": sigma_log__}
    normal_model_logp = normal_model_with_censored_logp(points)
    expected_model_logp = expected_normal_model_logp(points)
    np.testing.assert_allclose(normal_model_logp, expected_model_logp)


@pytest.mark.parametrize(
    "mu",
    [
        0,
        np.arange(10),
    ],
    ids=["scalar", "vector"],
)
def test_censored_logp(mu) -> None:
    n_points = 10
    observed = np.zeros(n_points)
    coords = {"idx": range(n_points)}
    with pm.Model(coords=coords) as model:
        normal = Prior("Normal", dims="idx")
        Censored(normal, lower=0).create_likelihood_variable(
            "censored_normal",
            observed=observed,
            mu=mu,
        )
    logp = model.compile_logp()

    with pm.Model() as expected_model:
        pm.Censored(
            "censored_normal",
            pm.Normal.dist(mu=mu, sigma=1, shape=n_points),
            lower=0,
            upper=np.inf,
            observed=observed,
        )
    expected_logp = expected_model.compile_logp()

    point = {}
    np.testing.assert_allclose(logp(point), expected_logp(point))


def test_scaled_initializes_correctly() -> None:
    """Test that the Scaled class initializes correctly."""
    normal = Prior("Normal", mu=0, sigma=1)
    scaled = Scaled(normal, factor=2.0)

    assert scaled.dist == normal
    assert scaled.factor == 2.0


def test_scaled_dims_property() -> None:
    """Test that the dims property returns the dimensions of the underlying distribution."""
    normal = Prior("Normal", mu=0, sigma=1, dims="channel")
    scaled = Scaled(normal, factor=2.0)

    assert scaled.dims == ("channel",)

    # Test with multiple dimensions
    normal.dims = ("channel", "geo")
    assert scaled.dims == ("channel", "geo")


def test_scaled_create_variable() -> None:
    """Test that the create_variable method properly scales the variable."""
    normal = Prior("Normal", mu=0, sigma=1)
    scaled = Scaled(normal, factor=2.0)

    with pm.Model() as model:
        scaled_var = scaled.create_variable("scaled_var")

    # Check that both the scaled and unscaled variables exist
    assert "scaled_var" in model
    assert "scaled_var_unscaled" in model

    # The deterministic node should be the scaled variable
    assert model["scaled_var"] == scaled_var


def test_scaled_creates_correct_dimensions() -> None:
    """Test that the scaled variable has the correct dimensions."""
    normal = Prior("Normal", dims="channel")
    scaled = Scaled(normal, factor=2.0)

    coords = {"channel": ["A", "B", "C"]}
    with pm.Model(coords=coords):
        scaled_var = scaled.create_variable("scaled_var")

    # Check that the scaled variable has the correct dimensions
    assert fast_eval(scaled_var).shape == (3,)


def test_scaled_applies_factor() -> None:
    """Test that the scaling factor is correctly applied."""
    normal = Prior("Normal", mu=0, sigma=1)
    factor = 3.5
    scaled = Scaled(normal, factor=factor)

    # Sample from prior to verify scaling
    prior = sample_prior(scaled, draws=10, name="scaled_var")
    df_prior = prior.to_dataframe()

    # Check that scaled values are original values times the factor
    unscaled_values = df_prior["scaled_var_unscaled"].to_numpy()
    scaled_values = df_prior["scaled_var"].to_numpy()
    np.testing.assert_allclose(scaled_values, unscaled_values * factor)


def test_scaled_with_tensor_factor() -> None:
    """Test that the Scaled class works with a tensor factor."""
    normal = Prior("Normal", mu=0, sigma=1)
    factor = pt.as_tensor_variable(2.5)
    scaled = Scaled(normal, factor=factor)

    # Sample from prior to verify tensor scaling
    prior = sample_prior(scaled, draws=10, name="scaled_var")
    df_prior = prior.to_dataframe()

    # Check that scaled values are original values times the factor
    unscaled_values = df_prior["scaled_var_unscaled"].to_numpy()
    scaled_values = df_prior["scaled_var"].to_numpy()
    np.testing.assert_allclose(scaled_values, unscaled_values * 2.5)


def test_scaled_with_hierarchical_prior() -> None:
    """Test that the Scaled class works with hierarchical priors."""
    normal = Prior("Normal", mu=Prior("Normal"), sigma=Prior("HalfNormal"), dims="channel")
    scaled = Scaled(normal, factor=2.0)

    coords = {"channel": ["A", "B", "C"]}
    with pm.Model(coords=coords) as model:
        scaled.create_variable("scaled_var")

    # Check that all necessary variables were created
    assert "scaled_var" in model
    assert "scaled_var_unscaled" in model
    assert "scaled_var_unscaled_mu" in model
    assert "scaled_var_unscaled_sigma" in model


def test_scaled_sample_prior() -> None:
    """Test that sample_prior works with the Scaled class."""
    normal = Prior("Normal", dims="channel")
    scaled = Scaled(normal, factor=2.0)

    coords = {"channel": ["A", "B", "C"]}
    prior = sample_prior(scaled, coords=coords, draws=25, name="scaled_var")

    assert isinstance(prior, xr.Dataset)
    assert prior.sizes == {"chain": 1, "draw": 25, "channel": 3}
    assert "scaled_var" in prior
    assert "scaled_var_unscaled" in prior


def test_getattr() -> None:
    assert pr.Normal() == Prior("Normal")


def test_import_directly() -> None:
    try:
        from pymc_extras.prior import Normal
    except Exception as e:
        pytest.fail(f"Unexpected exception: {e}")

    assert Normal() == Prior("Normal")


def test_import_incorrect_directly() -> None:
    match = "PyMC doesn't have a distribution of name 'SomeIncorrectDistribution'"
    with pytest.raises(UnsupportedDistributionError, match=match):
        from pymc_extras.prior import SomeIncorrectDistribution  # noqa: F401


@pytest.fixture
def alternative_prior_deserialize():
    def is_type(data):
        return isinstance(data, dict) and "distribution" in data

    def deserialize(data):
        return Prior(**data)

    register_deserialization(is_type=is_type, deserialize=deserialize)

    yield

    DESERIALIZERS.pop()


def test_censored_with_alternative(alternative_prior_deserialize) -> None:
    data = {
        "class": "Censored",
        "data": {
            "dist": {
                "distribution": "Normal",
            },
            "lower": 0,
            "upper": 10,
        },
    }

    instance = deserialize(data)

    assert isinstance(instance, Censored)
    assert instance.lower == 0
    assert instance.upper == 10
    assert instance.distribution == Prior("Normal")



import numpy as np
import pymc as pm
import pytensor
import pytensor.tensor as pt

from pymc_extras.statespace.core import PyMCStateSpace
from pymc_extras.statespace.filters.distributions import LinearGaussianStateSpace
from pymc_extras.statespace.utils.constants import SHORT_NAME_TO_LONG


def compile_statespace(
    statespace_model: PyMCStateSpace, steps: int | None = None, **compile_kwargs
):
    if steps is None:
        steps = pt.iscalar("steps")

    x0, _, c, d, T, Z, R, H, Q = statespace_model._unpack_statespace_with_placeholders()

    sequence_names = [x.name for x in [c, d] if x.ndim == 2]
    sequence_names += [x.name for x in [T, Z, R, H, Q] if x.ndim == 3]

    rename_dict = {v: k for k, v in SHORT_NAME_TO_LONG.items()}
    sequence_names = list(map(rename_dict.get, sequence_names))

    P0 = pt.zeros((x0.shape[0], x0.shape[0]))

    outputs = LinearGaussianStateSpace.dist(
        x0, P0, c, d, T, Z, R, H, Q, steps=steps, sequence_names=sequence_names
    )

    inputs = list(pytensor.graph.basic.explicit_graph_inputs(outputs))

    _f = pm.compile(inputs, outputs, on_unused_input="ignore", **compile_kwargs)

    def f(*, draws=1, **params):
        if isinstance(steps, pt.Variable):
            inner_steps = params.get("steps", 100)
        else:
            inner_steps = steps

        output = [np.empty((draws, inner_steps + 1, x.type.shape[-1])) for x in outputs]
        for i in range(draws):
            draw = _f(**params)
            for j, x in enumerate(draw):
                output[j][i] = x
        return [x.squeeze() for x in output]

    return f



import numpy as np

from pytensor import tensor as pt
from pytensor.tensor.slinalg import block_diag

from pymc_extras.statespace.models.structural.core import Component
from pymc_extras.statespace.models.structural.utils import _frequency_transition_block


class CycleComponent(Component):
    r"""
    A component for modeling longer-term cyclical effects

    Supports both univariate and multivariate time series. For multivariate time series,
    each endogenous variable gets its own independent cycle component with separate
    cosine/sine states and optional variable-specific innovation variances.

    Parameters
    ----------
    name: str
        Name of the component. Used in generated coordinates and state names. If None, a descriptive name will be
        used.

    cycle_length: int, optional
        The length of the cycle, in the calendar units of your data. For example, if your data is monthly, and you
        want to model a 12-month cycle, use ``cycle_length=12``. You cannot specify both ``cycle_length`` and
        ``estimate_cycle_length``.

    estimate_cycle_length: bool, default False
        Whether to estimate the cycle length. If True, an additional parameter, ``cycle_length`` will be added to the
        model. You cannot specify both ``cycle_length`` and ``estimate_cycle_length``.

    dampen: bool, default False
        Whether to dampen the cycle by multiplying by a dampening factor :math:`\rho` at every timestep. If true,
        an additional parameter, ``dampening_factor`` will be added to the model.

    innovations: bool, default True
        Whether to include stochastic innovations in the strength of the seasonal effect. If True, an additional
        parameter, ``sigma_{name}`` will be added to the model.
        For multivariate time series, this is a vector (variable-specific innovation variances).

    observed_state_names: list[str], optional
        Names of the observed state variables. For univariate time series, defaults to ``["data"]``.
        For multivariate time series, specify a list of names for each endogenous variable.

    share_states: bool, default False
        Whether latent states are shared across the observed states. If True, there will be only one set of latent
        states, which are observed by all observed states. If False, each observed state has its own set of
        latent states. This argument has no effect if `k_endog` is 1.

    Notes
    -----
    The cycle component is very similar in implementation to the frequency domain seasonal component, expect that it
    is restricted to n=1. The cycle component can be expressed:

    .. math::
        \begin{align}
            \gamma_t &= \rho \gamma_{t-1} \cos \lambda + \rho \gamma_{t-1}^\star \sin \lambda + \omega_{t} \\
            \gamma_{t}^\star &= -\rho \gamma_{t-1} \sin \lambda + \rho \gamma_{t-1}^\star \cos \lambda + \omega_{t}^\star \\
            \lambda &= \frac{2\pi}{s}
        \end{align}

    Where :math:`s` is the ``cycle_length``. [1] recommend that this component be used for longer term cyclical
    effects, such as business cycles, and that the seasonal component be used for shorter term effects, such as
    weekly or monthly seasonality.

    Unlike a FrequencySeasonality component, the length of a CycleComponent can be estimated.

    **Multivariate Support:**
    For multivariate time series with k endogenous variables, the component creates:
    - 2k states (cosine and sine components for each variable)
    - Block diagonal transition and selection matrices
    - Variable-specific innovation variances (optional)
    - Proper parameter shapes: (k, 2) for initial states, (k,) for innovation variances

    Examples
    --------
    **Univariate Example:**
    Estimate a business cycle with length between 6 and 12 years:

    .. code:: python

        from pymc_extras.statespace import structural as st
        import pymc as pm
        import pytensor.tensor as pt
        import pandas as pd
        import numpy as np

        data = np.random.normal(size=(100, 1))

        # Build the structural model
        grw = st.LevelTrendComponent(order=1, innovations_order=1)
        cycle = st.CycleComponent(
            "business_cycle", cycle_length=12, estimate_cycle_length=False, innovations=True, dampen=True
        )
        ss_mod = (grw + cycle).build()

        # Estimate with PyMC
        with pm.Model(coords=ss_mod.coords) as model:
            P0 = pm.Deterministic('P0', pt.eye(ss_mod.k_states), dims=ss_mod.param_dims['P0'])

            initial_level_trend = pm.Normal('initial_level_trend', dims=ss_mod.param_dims['initial_level_trend'])
            sigma_level_trend = pm.HalfNormal('sigma_level_trend', dims=ss_mod.param_dims['sigma_level_trend'])

            business_cycle = pm.Normal("business_cycle", dims=ss_mod.param_dims["business_cycle"])
            dampening = pm.Beta("dampening_factor_business_cycle", 2, 2)
            sigma_cycle = pm.HalfNormal("sigma_business_cycle", sigma=1)

            ss_mod.build_statespace_graph(data)
            idata = pm.sample(
                nuts_sampler="nutpie", nuts_sampler_kwargs={"backend": "JAX", "gradient_backend": "JAX"}
            )

    **Multivariate Example:**
    Model cycles for multiple economic indicators with variable-specific innovation variances:

    .. code:: python

        # Multivariate cycle component
        cycle = st.CycleComponent(
            name='business_cycle',
            cycle_length=12,
            estimate_cycle_length=False,
            innovations=True,
            dampen=True,
            observed_state_names=['gdp', 'unemployment', 'inflation']
        )
        ss_mod = cycle.build()

        with pm.Model(coords=ss_mod.coords) as model:
            P0 = pm.Deterministic("P0", pt.eye(ss_mod.k_states), dims=ss_mod.param_dims["P0"])
            # Initial states: shape (3, 2) for 3 variables, 2 states each
            business_cycle = pm.Normal('business_cycle', dims=ss_mod.param_dims["business_cycle"])

            # Dampening factor: scalar (shared across variables)
            dampening = pm.Beta("dampening_factor_business_cycle", 2, 2)

            # Innovation variances: shape (3,) for variable-specific variances
            sigma_cycle = pm.HalfNormal(
                "sigma_business_cycle", dims=ss_mod.param_dims["sigma_business_cycle"]
            )

            ss_mod.build_statespace_graph(data)
            idata = pm.sample(
                nuts_sampler="nutpie", nuts_sampler_kwargs={"backend": "JAX", "gradient_backend": "JAX"}
            )

    References
    ----------
    .. [1] Durbin, James, and Siem Jan Koopman. 2012.
        Time Series Analysis by State Space Methods: Second Edition.
        Oxford University Press.
    """

    def __init__(
        self,
        name: str | None = None,
        cycle_length: int | None = None,
        estimate_cycle_length: bool = False,
        dampen: bool = False,
        innovations: bool = True,
        observed_state_names: list[str] | None = None,
        share_states: bool = False,
    ):
        if observed_state_names is None:
            observed_state_names = ["data"]

        if cycle_length is None and not estimate_cycle_length:
            raise ValueError("Must specify cycle_length if estimate_cycle_length is False")
        if cycle_length is not None and estimate_cycle_length:
            raise ValueError("Cannot specify cycle_length if estimate_cycle_length is True")
        if name is None:
            cycle = int(cycle_length) if cycle_length is not None else "Estimate"
            name = f"Cycle[s={cycle}, dampen={dampen}, innovations={innovations}]"

        self.share_states = share_states
        self.estimate_cycle_length = estimate_cycle_length
        self.cycle_length = cycle_length
        self.innovations = innovations
        self.dampen = dampen
        self.n_coefs = 1

        k_endog = len(observed_state_names)

        k_states = 2 if share_states else 2 * k_endog
        k_posdef = 2 if share_states else 2 * k_endog

        obs_state_idx = np.zeros(k_states)
        obs_state_idx[slice(0, k_states, 2)] = 1

        super().__init__(
            name=name,
            k_endog=k_endog,
            k_states=k_states,
            k_posdef=k_posdef,
            measurement_error=False,
            combine_hidden_states=True,
            obs_state_idxs=obs_state_idx,
            observed_state_names=observed_state_names,
            share_states=share_states,
        )

    def make_symbolic_graph(self) -> None:
        k_endog = self.k_endog
        k_endog_effective = 1 if self.share_states else k_endog

        Z = np.array([1.0, 0.0]).reshape((1, -1))
        design_matrix = block_diag(*[Z for _ in range(k_endog_effective)])
        self.ssm["design", :, :] = pt.as_tensor_variable(design_matrix)

        # selection matrix R defines structure of innovations (always identity for cycle components)
        # when innovations=False, state cov Q=0, hence R @ Q @ R.T = 0
        R = np.eye(2)  # 2x2 identity for each cycle component
        selection_matrix = block_diag(*[R for _ in range(k_endog_effective)])
        self.ssm["selection", :, :] = pt.as_tensor_variable(selection_matrix)

        init_state = self.make_and_register_variable(
            f"params_{self.name}",
            shape=(k_endog_effective, 2) if k_endog_effective > 1 else (self.k_states,),
        )
        self.ssm["initial_state", :] = init_state.ravel()

        if self.estimate_cycle_length:
            lamb = self.make_and_register_variable(f"length_{self.name}", shape=())
        else:
            lamb = self.cycle_length

        if self.dampen:
            rho = self.make_and_register_variable(f"dampening_factor_{self.name}", shape=())
        else:
            rho = 1

        T = rho * _frequency_transition_block(lamb, j=1)
        transition = block_diag(*[T for _ in range(k_endog_effective)])
        self.ssm["transition"] = pt.specify_shape(transition, (self.k_states, self.k_states))

        if self.innovations:
            if k_endog_effective == 1:
                sigma_cycle = self.make_and_register_variable(f"sigma_{self.name}", shape=())
                self.ssm["state_cov", :, :] = pt.eye(self.k_posdef) * sigma_cycle**2
            else:
                sigma_cycle = self.make_and_register_variable(
                    f"sigma_{self.name}", shape=(k_endog_effective,)
                )
                state_cov = block_diag(
                    *[pt.eye(2) * sigma_cycle[i] ** 2 for i in range(k_endog_effective)]
                )
                self.ssm["state_cov"] = pt.specify_shape(state_cov, (self.k_states, self.k_states))
        else:
            # explicitly set state cov to 0 when no innovations
            self.ssm["state_cov", :, :] = pt.zeros((self.k_posdef, self.k_posdef))

    def populate_component_properties(self):
        k_endog = self.k_endog
        k_endog_effective = 1 if self.share_states else k_endog

        base_names = [f"{f}_{self.name}" for f in ["Cos", "Sin"]]

        if self.share_states:
            self.state_names = [f"{name}[shared]" for name in base_names]
        else:
            self.state_names = [
                f"{name}[{var_name}]" if k_endog_effective > 1 else name
                for var_name in self.observed_state_names
                for name in base_names
            ]

        self.param_names = [f"params_{self.name}"]

        if k_endog_effective == 1:
            self.param_dims = {f"params_{self.name}": (f"state_{self.name}",)}
            self.coords = {f"state_{self.name}": base_names}
            self.param_info = {
                f"params_{self.name}": {
                    "shape": (2,),
                    "constraints": None,
                    "dims": (f"state_{self.name}",),
                }
            }
        else:
            self.param_dims = {f"params_{self.name}": (f"endog_{self.name}", f"state_{self.name}")}
            self.coords = {
                f"state_{self.name}": [f"Cos_{self.name}", f"Sin_{self.name}"],
                f"endog_{self.name}": self.observed_state_names,
            }
            self.param_info = {
                f"params_{self.name}": {
                    "shape": (k_endog_effective, 2),
                    "constraints": None,
                    "dims": (f"endog_{self.name}", f"state_{self.name}"),
                }
            }

        if self.estimate_cycle_length:
            self.param_names += [f"length_{self.name}"]
            self.param_info[f"length_{self.name}"] = {
                "shape": () if k_endog_effective == 1 else (k_endog_effective,),
                "constraints": "Positive, non-zero",
                "dims": None if k_endog_effective == 1 else (f"endog_{self.name}",),
            }

        if self.dampen:
            self.param_names += [f"dampening_factor_{self.name}"]
            self.param_info[f"dampening_factor_{self.name}"] = {
                "shape": () if k_endog_effective == 1 else (k_endog_effective,),
                "constraints": "0 < x ≤ 1",
                "dims": None if k_endog_effective == 1 else (f"endog_{self.name}",),
            }

        if self.innovations:
            self.param_names += [f"sigma_{self.name}"]
            if k_endog_effective == 1:
                self.param_info[f"sigma_{self.name}"] = {
                    "shape": (),
                    "constraints": "Positive",
                    "dims": None,
                }
            else:
                self.param_dims[f"sigma_{self.name}"] = (f"endog_{self.name}",)
                self.param_info[f"sigma_{self.name}"] = {
                    "shape": (k_endog_effective,),
                    "constraints": "Positive",
                    "dims": (f"endog_{self.name}",),
                }
            self.shock_names = self.state_names.copy()



import pytensor.tensor as pt
import pytest

from pymc.distributions import CustomDist
from pymc.variational.minibatch_rv import create_minibatch_rv
from pytensor.tensor.type_other import NoneTypeT

from pymc_extras.model.marginal.graph_analysis import (
    is_conditional_dependent,
    subgraph_batch_dim_connection,
)


def test_is_conditional_dependent_static_shape():
    """Test that we don't consider dependencies through "constant" shape Ops"""
    x1 = pt.matrix("x1", shape=(None, 5))
    y1 = pt.random.normal(size=pt.shape(x1))
    assert is_conditional_dependent(y1, x1, [x1, y1])

    x2 = pt.matrix("x2", shape=(9, 5))
    y2 = pt.random.normal(size=pt.shape(x2))
    assert not is_conditional_dependent(y2, x2, [x2, y2])


class TestSubgraphBatchDimConnection:
    def test_dimshuffle(self):
        inp = pt.tensor(shape=(5, 1, 4, 3))
        out1 = pt.matrix_transpose(inp)
        out2 = pt.expand_dims(inp, 1)
        out3 = pt.squeeze(inp)
        [dims1, dims2, dims3] = subgraph_batch_dim_connection(inp, [out1, out2, out3])
        assert dims1 == (0, 1, 3, 2)
        assert dims2 == (0, None, 1, 2, 3)
        assert dims3 == (0, 2, 3)

    def test_careduce(self):
        inp = pt.tensor(shape=(4, 3, 2))

        out = pt.sum(inp[:, None], axis=(1,))
        [dims] = subgraph_batch_dim_connection(inp, [out])
        assert dims == (0, 1, 2)

        invalid_out = pt.sum(inp, axis=(1,))
        with pytest.raises(ValueError, match="Use of known dimensions"):
            subgraph_batch_dim_connection(inp, [invalid_out])

    def test_subtensor(self):
        inp = pt.tensor(shape=(4, 3, 2))

        invalid_out = inp[0, :1]
        with pytest.raises(
            ValueError,
            match="Partial slicing or indexing of known dimensions not supported",
        ):
            subgraph_batch_dim_connection(inp, [invalid_out])

        # If we are selecting dummy / unknown dimensions that's fine
        valid_out = pt.expand_dims(inp, (0, 1))[0, :1]
        [dims] = subgraph_batch_dim_connection(inp, [valid_out])
        assert dims == (None, 0, 1, 2)

    def test_advanced_subtensor_value(self):
        inp = pt.tensor(shape=(2, 4))
        intermediate_out = inp[:, None, :, None] + pt.zeros((2, 3, 4, 5))

        # Index on an unlabled dim introduced by broadcasting with zeros
        out = intermediate_out[:, [0, 0, 1, 2]]
        [dims] = subgraph_batch_dim_connection(inp, [out])
        assert dims == (0, None, 1, None)

        # Indexing that introduces more dimensions
        out = intermediate_out[:, [[0, 0], [1, 2]], :]
        [dims] = subgraph_batch_dim_connection(inp, [out])
        assert dims == (0, None, None, 1, None)

        # Special case where advanced dims are moved to the front of the output
        out = intermediate_out[:, [0, 0, 1, 2], :, 0]
        [dims] = subgraph_batch_dim_connection(inp, [out])
        assert dims == (None, 0, 1)

        # Indexing on a labeled dim fails
        out = intermediate_out[:, :, [0, 0, 1, 2]]
        with pytest.raises(ValueError, match="Partial slicing or advanced integer indexing"):
            subgraph_batch_dim_connection(inp, [out])

    def test_advanced_subtensor_key(self):
        inp = pt.tensor(shape=(5, 5), dtype=int)
        base = pt.zeros((2, 3, 4))

        out = base[inp]
        [dims] = subgraph_batch_dim_connection(inp, [out])
        assert dims == (0, 1, None, None)

        out = base[:, :, inp]
        [dims] = subgraph_batch_dim_connection(inp, [out])
        assert dims == (
            None,
            None,
            0,
            1,
        )

        out = base[1:, 0, inp]
        [dims] = subgraph_batch_dim_connection(inp, [out])
        assert dims == (None, 0, 1)

        # Special case where advanced dims are moved to the front of the output
        out = base[0, :, inp]
        [dims] = subgraph_batch_dim_connection(inp, [out])
        assert dims == (0, 1, None)

        # Mix keys dimensions
        out = base[:, inp, inp.T]
        with pytest.raises(ValueError, match="Different known dimensions mixed via broadcasting"):
            subgraph_batch_dim_connection(inp, [out])

    def test_elemwise(self):
        inp = pt.tensor(shape=(5, 5))

        out = inp + inp
        [dims] = subgraph_batch_dim_connection(inp, [out])
        assert dims == (0, 1)

        out = inp + inp.T
        with pytest.raises(ValueError, match="Different known dimensions mixed via broadcasting"):
            subgraph_batch_dim_connection(inp, [out])

        out = inp[None, :, None, :] + inp[:, None, :, None]
        with pytest.raises(
            ValueError, match="Same known dimension used in different axis after broadcasting"
        ):
            subgraph_batch_dim_connection(inp, [out])

    def test_blockwise(self):
        inp = pt.tensor(shape=(5, 4))

        invalid_out = inp @ pt.ones((4, 3))
        with pytest.raises(ValueError, match="Use of known dimensions"):
            subgraph_batch_dim_connection(inp, [invalid_out])

        out = (inp[:, :, None, None] + pt.zeros((2, 3))) @ pt.ones((2, 3))
        [dims] = subgraph_batch_dim_connection(inp, [out])
        assert dims == (0, 1, None, None)

    def test_random_variable(self):
        inp = pt.tensor(shape=(5, 4, 3))

        out1 = pt.random.normal(loc=inp)
        out2 = pt.random.categorical(p=inp[..., None])
        out3 = pt.random.multivariate_normal(mean=inp[..., None], cov=pt.eye(1))
        [dims1, dims2, dims3] = subgraph_batch_dim_connection(inp, [out1, out2, out3])
        assert dims1 == (0, 1, 2)
        assert dims2 == (0, 1, 2)
        assert dims3 == (0, 1, 2, None)

        invalid_out = pt.random.categorical(p=inp)
        with pytest.raises(ValueError, match="Use of known dimensions"):
            subgraph_batch_dim_connection(inp, [invalid_out])

        invalid_out = pt.random.multivariate_normal(mean=inp, cov=pt.eye(3))
        with pytest.raises(ValueError, match="Use of known dimensions"):
            subgraph_batch_dim_connection(inp, [invalid_out])

    def test_minibatched_random_variable(self):
        inp = pt.tensor(shape=(4, 3, 2))
        out1 = pt.random.normal(loc=inp)
        out2 = create_minibatch_rv(out1, total_size=(10, 10, 10))
        [dims1] = subgraph_batch_dim_connection(inp, [out2])
        assert dims1 == (0, 1, 2)

    def test_symbolic_random_variable(self):
        inp = pt.tensor(shape=(4, 3, 2))

        # Test univariate
        out = CustomDist.dist(
            inp,
            dist=lambda mu, size: pt.random.normal(loc=mu, size=size),
        )
        [dims] = subgraph_batch_dim_connection(inp, [out])
        assert dims == (0, 1, 2)

        # Test multivariate
        def dist(mu, size):
            if isinstance(size.type, NoneTypeT):
                size = mu.shape
            return pt.random.normal(loc=mu[..., None], size=(*size, 2))

        out = CustomDist.dist(inp, dist=dist, size=(4, 3, 2), signature="()->(2)")
        [dims] = subgraph_batch_dim_connection(inp, [out])
        assert dims == (0, 1, 2, None)



import numpy as np
import pytest

from pymc_extras.statespace.utils.constants import (
    ALL_STATE_AUX_DIM,
    ALL_STATE_DIM,
    OBS_STATE_AUX_DIM,
    OBS_STATE_DIM,
    SHOCK_AUX_DIM,
    SHOCK_DIM,
)

TEST_SEED = sum(map(ord, "Structural Statespace"))


@pytest.fixture(scope="session")
def rng():
    return np.random.default_rng(TEST_SEED)


def _assert_basic_coords_correct(mod):
    assert mod.coords[ALL_STATE_DIM] == mod.state_names
    assert mod.coords[ALL_STATE_AUX_DIM] == mod.state_names
    assert mod.coords[SHOCK_DIM] == mod.shock_names
    assert mod.coords[SHOCK_AUX_DIM] == mod.shock_names
    expected_obs = mod.observed_state_names if hasattr(mod, "observed_state_names") else ["data"]

    assert mod.coords[OBS_STATE_DIM] == expected_obs
    assert mod.coords[OBS_STATE_AUX_DIM] == expected_obs



from pymc_extras.statespace.models import structural
from pymc_extras.statespace.models.ETS import BayesianETS
from pymc_extras.statespace.models.SARIMAX import BayesianSARIMAX
from pymc_extras.statespace.models.VARMAX import BayesianVARMAX

__all__ = ["BayesianSARIMAX", "BayesianVARMAX", "BayesianETS", "structural"]



import numpy as np

from pytensor import tensor as pt

from pymc_extras.statespace.models.structural.core import Component
from pymc_extras.statespace.utils.constants import TIME_DIM


class RegressionComponent(Component):
    r"""
    Regression component for exogenous variables in a structural time series model

    Parameters
    ----------
    k_exog : int | None, default None
        Number of exogenous variables to include in the regression. Must be specified if
        state_names is not provided.

    name : str | None, default "regression"
        A name for this regression component. Used to label dimensions and coordinates.

    state_names : list[str] | None, default None
        List of strings for regression coefficient labels. If provided, must be of length
        k_exog. If None and k_exog is provided, coefficients will be named
        "{name}_1, {name}_2, ...".

    observed_state_names : list[str] | None, default None
        List of strings for observed state labels. If None, defaults to ["data"].

    innovations : bool, default False
        Whether to include stochastic innovations in the regression coefficients,
        allowing them to vary over time. If True, coefficients follow a random walk.

    share_states: bool, default False
        Whether latent states are shared across the observed states. If True, there will be only one set of latent
        states, which are observed by all observed states. If False, each observed state has its own set of
        latent states.

    Notes
    -----
    This component implements regression with exogenous variables in a structural time series
    model. The regression component can be expressed as:

    .. math::
        y_t = \beta_t^T x_t + \epsilon_t

    Where :math:`y_t` is the dependent variable, :math:`x_t` is the vector of exogenous
    variables, :math:`\beta_t` is the vector of regression coefficients, and :math:`\epsilon_t`
    is the error term.

    When ``innovations=False`` (default), the coefficients are constant over time:
    :math:`\beta_t = \beta_0` for all t.

    When ``innovations=True``, the coefficients follow a random walk:
    :math:`\beta_{t+1} = \beta_t + \eta_t`, where :math:`\eta_t \sim N(0, \Sigma_\beta)`.

    The component supports both univariate and multivariate regression. In the multivariate
    case, separate coefficients are estimated for each endogenous variable (i.e time series).

    Examples
    --------
    Simple regression with constant coefficients:

    .. code:: python

        from pymc_extras.statespace import structural as st
        import pymc as pm
        import pytensor.tensor as pt

        trend = st.LevelTrendComponent(order=1, innovations_order=1)
        regression = st.RegressionComponent(k_exog=2, state_names=['intercept', 'slope'])
        ss_mod = (trend + regression).build()

        with pm.Model(coords=ss_mod.coords) as model:
            # Prior for regression coefficients
            betas = pm.Normal('betas', dims=ss_mod.param_dims['beta_regression'])

            # Prior for trend innovations
            sigma_trend = pm.Exponential('sigma_trend', 1)

            ss_mod.build_statespace_graph(data)
            idata = pm.sample()

    Multivariate regression with time-varying coefficients:
    - There are 2 exogenous variables (price and income effects)
    - There are 2 endogenous variables (sales and revenue)
    - The regression coefficients are allowed to vary over time (`innovations=True`)

    .. code:: python

        regression = st.RegressionComponent(
            k_exog=2,
            state_names=['price_effect', 'income_effect'],
            observed_state_names=['sales', 'revenue'],
            innovations=True
        )

        with pm.Model(coords=ss_mod.coords) as model:
            betas = pm.Normal('betas', dims=ss_mod.param_dims['beta_regression'])

            # Innovation variance for time-varying coefficients
            sigma_beta = pm.Exponential('sigma_beta', 1, dims=ss_mod.param_dims['sigma_beta_regression'])

            ss_mod.build_statespace_graph(data)
            idata = pm.sample()
    """

    def __init__(
        self,
        k_exog: int | None = None,
        name: str | None = "regression",
        state_names: list[str] | None = None,
        observed_state_names: list[str] | None = None,
        innovations=False,
        share_states: bool = False,
    ):
        self.share_states = share_states

        if observed_state_names is None:
            observed_state_names = ["data"]

        self.innovations = innovations
        k_exog = self._handle_input_data(k_exog, state_names, name)

        k_states = k_exog
        k_endog = len(observed_state_names)
        k_posdef = k_exog

        super().__init__(
            name=name,
            k_endog=k_endog,
            k_states=k_states * k_endog if not share_states else k_states,
            k_posdef=k_posdef * k_endog if not share_states else k_posdef,
            state_names=self.state_names,
            share_states=share_states,
            observed_state_names=observed_state_names,
            measurement_error=False,
            combine_hidden_states=False,
            exog_names=[f"data_{name}"],
            obs_state_idxs=np.ones(k_states),
        )

    @staticmethod
    def _get_state_names(k_exog: int | None, state_names: list[str] | None, name: str):
        if k_exog is None and state_names is None:
            raise ValueError("Must specify at least one of k_exog or state_names")
        if state_names is not None and k_exog is not None:
            if len(state_names) != k_exog:
                raise ValueError(f"Expected {k_exog} state names, found {len(state_names)}")
        elif k_exog is None:
            k_exog = len(state_names)
        else:
            state_names = [f"{name}_{i + 1}" for i in range(k_exog)]

        return k_exog, state_names

    def _handle_input_data(self, k_exog: int, state_names: list[str] | None, name) -> int:
        k_exog, state_names = self._get_state_names(k_exog, state_names, name)
        self.state_names = state_names

        return k_exog

    def make_symbolic_graph(self) -> None:
        k_endog = self.k_endog
        k_endog_effective = 1 if self.share_states else k_endog

        k_states = self.k_states // k_endog_effective

        betas = self.make_and_register_variable(
            f"beta_{self.name}", shape=(k_endog, k_states) if k_endog_effective > 1 else (k_states,)
        )
        regression_data = self.make_and_register_data(f"data_{self.name}", shape=(None, k_states))

        self.ssm["initial_state", :] = betas.ravel()
        self.ssm["transition", :, :] = pt.eye(self.k_states)
        self.ssm["selection", :, :] = pt.eye(self.k_states)

        if self.share_states:
            self.ssm["design"] = pt.specify_shape(
                pt.join(1, *[pt.expand_dims(regression_data, 1) for _ in range(k_endog)]),
                (None, k_endog, self.k_states),
            )
        else:
            Z = pt.linalg.block_diag(*[pt.expand_dims(regression_data, 1) for _ in range(k_endog)])
            self.ssm["design"] = pt.specify_shape(
                Z, (None, k_endog, regression_data.type.shape[1] * k_endog)
            )

        if self.innovations:
            sigma_beta = self.make_and_register_variable(
                f"sigma_beta_{self.name}",
                (k_states,) if k_endog_effective == 1 else (k_endog, k_states),
            )
            row_idx, col_idx = np.diag_indices(self.k_states)
            self.ssm["state_cov", row_idx, col_idx] = sigma_beta.ravel() ** 2

    def populate_component_properties(self) -> None:
        k_endog = self.k_endog
        k_endog_effective = 1 if self.share_states else k_endog

        k_states = self.k_states // k_endog_effective

        if self.share_states:
            self.shock_names = [f"{state_name}_shared" for state_name in self.state_names]
        else:
            self.shock_names = self.state_names

        self.param_names = [f"beta_{self.name}"]
        self.data_names = [f"data_{self.name}"]
        self.param_dims = {
            f"beta_{self.name}": (f"endog_{self.name}", f"state_{self.name}")
            if k_endog_effective > 1
            else (f"state_{self.name}",)
        }

        base_names = self.state_names

        if self.share_states:
            self.state_names = [f"{name}[{self.name}_shared]" for name in base_names]
        else:
            self.state_names = [
                f"{name}[{obs_name}]"
                for obs_name in self.observed_state_names
                for name in base_names
            ]

        self.param_info = {
            f"beta_{self.name}": {
                "shape": (k_endog_effective, k_states) if k_endog_effective > 1 else (k_states,),
                "constraints": None,
                "dims": (f"endog_{self.name}", f"state_{self.name}")
                if k_endog_effective > 1
                else (f"state_{self.name}",),
            },
        }

        self.data_info = {
            f"data_{self.name}": {
                "shape": (None, k_states),
                "dims": (TIME_DIM, f"state_{self.name}"),
            },
        }
        self.coords = {
            f"state_{self.name}": base_names,
            f"endog_{self.name}": self.observed_state_names,
        }

        if self.innovations:
            self.param_names += [f"sigma_beta_{self.name}"]
            self.param_dims[f"sigma_beta_{self.name}"] = (f"state_{self.name}",)
            self.param_info[f"sigma_beta_{self.name}"] = {
                "shape": (k_states,),
                "constraints": "Positive",
                "dims": (f"state_{self.name}",)
                if k_endog_effective == 1
                else (f"endog_{self.name}", f"state_{self.name}"),
            }



from abc import ABC

import numpy as np
import pytensor
import pytensor.tensor as pt

from pymc.pytensorf import constant_fold
from pytensor.graph.basic import Variable
from pytensor.raise_op import Assert
from pytensor.tensor import TensorVariable
from pytensor.tensor.slinalg import solve_triangular

from pymc_extras.statespace.filters.utilities import (
    quad_form_sym,
    split_vars_into_seq_and_nonseq,
    stabilize,
)
from pymc_extras.statespace.utils.constants import (
    FILTER_OUTPUT_NAMES,
    JITTER_DEFAULT,
    MATRIX_NAMES,
    MISSING_FILL,
)

MVN_CONST = pt.log(2 * pt.constant(np.pi, dtype="float64"))
PARAM_NAMES = MATRIX_NAMES[2:]

assert_time_varying_dim_correct = Assert(
    "The first dimension of a time varying matrix (the time dimension) must be "
    "equal to the first dimension of the data (the time dimension)."
)


class BaseFilter(ABC):
    def __init__(self):
        """
        Kalman Filter.

        Notes
        -----
        The BaseFilter class is an abstract base class (ABC) for implementing kalman filters.
        It defines common attributes and methods used by kalman filter implementations.

        Attributes
        ----------
        seq_names : list[str]
            A list of name representing time-varying statespace matrices. That is, inputs that will need to be
            provided to the `sequences` argument of `pytensor.scan`

        non_seq_names : list[str]
            A list of names representing static statespace matrices. That is, inputs that will need to be provided
            to the `non_sequences` argument of `pytensor.scan`
        """

        self.seq_names: list[str] = []
        self.non_seq_names: list[str] = []

        self.n_states = None
        self.n_posdef = None
        self.n_endog = None

        self.missing_fill_value: float | None = None
        self.cov_jitter = None

    def check_params(self, data, a0, P0, c, d, T, Z, R, H, Q):
        """
        Apply any checks on validity of inputs. For most filters this is just the identity function.
        """
        return data, a0, P0, c, d, T, Z, R, H, Q

    @staticmethod
    def add_check_on_time_varying_shapes(
        data: TensorVariable, sequence_params: list[TensorVariable]
    ) -> list[Variable]:
        """
        Insert a check that time-varying matrices match the data shape to the computational graph.

        If any matrices are time-varying, they need to have the same length as the data. This function wraps each
        element of `sequence_params` in an assert `Op` that makes sure all inputs have the correct shape.

        Parameters
        ----------
        data : TensorVariable
            The tensor representing the data.

        sequence_params : list[TensorVariable]
            A list of tensors to be provided to `pytensor.scan` as `sequences`.

        Returns
        -------
        list[TensorVariable]
            A list of tensors wrapped in an `Assert` `Op` that checks the shape of the 0th dimension on each is equal
             to the shape of the 0th dimension on the data.
        """
        # TODO: The PytensorRepresentation object puts the time dimension last, should the reshaping happen here in
        #    the Kalman filter, or in the StateSpaceModel, before passing into the KF?

        params_with_assert = [
            assert_time_varying_dim_correct(param, pt.eq(param.shape[0], data.shape[0]))
            for param in sequence_params
        ]

        return params_with_assert

    def unpack_args(self, args) -> tuple:
        """
        The order of inputs to the inner scan function is not known, since some, all, or none of the input matrices
        can be time varying. The order arguments are fed to the inner function is sequences, outputs_info,
        non-sequences. This function works out which matrices are where, and returns a standardized order expected
        by the kalman_step function.

        The standard order is: y, a0, P0, c, d, T, Z, R, H, Q
        """
        # If there are no sequence parameters (all params are static),
        # no changes are needed, params will be in order.
        args = list(args)
        n_seq = len(self.seq_names)
        if n_seq == 0:
            return tuple(args)

        # The first arg is always y
        y = args.pop(0)

        # There are always two outputs_info wedged between the seqs and non_seqs
        seqs, (a0, P0), non_seqs = args[:n_seq], args[n_seq : n_seq + 2], args[n_seq + 2 :]
        return_ordered = []
        for name in PARAM_NAMES:
            if name in self.seq_names:
                idx = self.seq_names.index(name)
                return_ordered.append(seqs[idx])
            else:
                idx = self.non_seq_names.index(name)
                return_ordered.append(non_seqs[idx])

        c, d, T, Z, R, H, Q = return_ordered

        return y, a0, P0, c, d, T, Z, R, H, Q

    def build_graph(
        self,
        data,
        a0,
        P0,
        c,
        d,
        T,
        Z,
        R,
        H,
        Q,
        return_updates=False,
        missing_fill_value=None,
        cov_jitter=None,
    ) -> list[TensorVariable] | tuple[list[TensorVariable], dict]:
        """
        Construct the computation graph for the Kalman filter. See [1] for details.

        Parameters
        ----------
        data : TensorVariable
            Data to be filtered

        return_updates: bool, default False
            Whether to return updates associated with the pytensor scan. Should only be requried to debug pruposes.

        missing_fill_value: float, default -9999
            Fill value used to mark missing values. Used to avoid PyMC's automatic interpolation, which conflict's with
            the Kalman filter's hidden state inference. Change if your data happens to have legitimate values of -9999

        cov_jitter: float, default 1e-8 or 1e-6 if pytensor.config.floatX is float32
            The Kalman filter is known to be numerically unstable, especially at half precision. This value is added to
            the diagonal of every covariance matrix -- predicted, filtered, and smoothed -- at every step, to ensure
            all matrices are strictly positive semi-definite.

            Obviously, if this can be zero, that's best. In general:
                - Having measurement error makes Kalman Filters more robust. A large source of numerical errors come
                  from the Filtered and Smoothed matrices having a zero in the (0, 0) position, which always occurs
                  when there is no measurement error.

                - The Univariate Filter is more robust than other filters, and can tolerate a lower jitter value

        References
        ----------
        .. [1] Koopman, Siem Jan, Neil Shephard, and Jurgen A. Doornik. 1999.
           Statistical Algorithms for Models in State Space Using SsfPack 2.2.
           Econometrics Journal 2 (1): 107-60. doi:10.1111/1368-423X.00023.
        """
        if missing_fill_value is None:
            missing_fill_value = MISSING_FILL
        if cov_jitter is None:
            cov_jitter = JITTER_DEFAULT

        self.missing_fill_value = missing_fill_value
        self.cov_jitter = cov_jitter

        [R_shape] = constant_fold([R.shape], raise_not_constant=False)
        [Z_shape] = constant_fold([Z.shape], raise_not_constant=False)

        self.n_states, self.n_shocks = R_shape[-2:]
        self.n_endog = Z_shape[-2]

        data, a0, P0, *params = self.check_params(data, a0, P0, c, d, T, Z, R, H, Q)

        sequences, non_sequences, seq_names, non_seq_names = split_vars_into_seq_and_nonseq(
            params, PARAM_NAMES
        )

        self.seq_names = seq_names
        self.non_seq_names = non_seq_names

        if len(sequences) > 0:
            sequences = self.add_check_on_time_varying_shapes(data, sequences)

        results, updates = pytensor.scan(
            self.kalman_step,
            sequences=[data, *sequences],
            outputs_info=[None, a0, None, None, P0, None, None],
            non_sequences=non_sequences,
            name="forward_kalman_pass",
            strict=False,
        )

        filter_results = self._postprocess_scan_results(results, a0, P0, n=data.type.shape[0])

        if return_updates:
            return filter_results, updates
        return filter_results

    def _postprocess_scan_results(self, results, a0, P0, n) -> list[TensorVariable]:
        """
        Transform the values returned by the Kalman Filter scan into a form expected by users. In particular:
        1. Append the initial state and covariance matrix to their respective Kalman predictions. This matches the
        output returned by Statsmodels state space models.

        2. Discard the last state and covariance matrix from the Kalman predictions. This is beacuse the kalman filter
        starts with the (random variable) initial state x0, and treats it as a predicted state. The first step (t=0)
        will filter x0 to make filtered_states[0], then do a predict step to make predicted_states[1]. This means
        the last step (t=T) predicted state will be a *forecast* for T+1. If the user wants this forecast, he should
        use the forecast method.

        3. Squeeze away extra dimensions from the filtered and predicted states, as well as the likelihoods.
        """
        (
            filtered_states,
            predicted_states,
            observed_states,
            filtered_covariances,
            predicted_covariances,
            observed_covariances,
            loglike_obs,
        ) = results

        predicted_states = pt.concatenate(
            [pt.expand_dims(a0, axis=(0,)), predicted_states[:-1]], axis=0
        )
        predicted_covariances = pt.concatenate(
            [pt.expand_dims(P0, axis=(0,)), predicted_covariances[:-1]], axis=0
        )

        filtered_states = pt.specify_shape(filtered_states, (n, self.n_states))
        filtered_states.name = FILTER_OUTPUT_NAMES[0]

        predicted_states = pt.specify_shape(predicted_states, (n, self.n_states))
        predicted_states.name = FILTER_OUTPUT_NAMES[1]

        filtered_covariances = pt.specify_shape(
            filtered_covariances, (n, self.n_states, self.n_states)
        )
        filtered_covariances.name = FILTER_OUTPUT_NAMES[2]

        predicted_covariances = pt.specify_shape(
            predicted_covariances, (n, self.n_states, self.n_states)
        )
        predicted_covariances.name = FILTER_OUTPUT_NAMES[3]

        observed_states = pt.specify_shape(observed_states, (n, self.n_endog))
        observed_states.name = FILTER_OUTPUT_NAMES[4]

        observed_covariances = pt.specify_shape(
            observed_covariances, (n, self.n_endog, self.n_endog)
        )
        observed_covariances.name = FILTER_OUTPUT_NAMES[5]

        loglike_obs = pt.specify_shape(loglike_obs.squeeze(), (n,))
        loglike_obs.name = "loglike_obs"

        filter_results = [
            filtered_states,
            predicted_states,
            observed_states,
            filtered_covariances,
            predicted_covariances,
            observed_covariances,
            loglike_obs,
        ]

        return filter_results

    def handle_missing_values(
        self, y, Z, H
    ) -> tuple[TensorVariable, TensorVariable, TensorVariable, float]:
        """
        Handle missing values in the observation data `y`

        Adjusts the design matrix `Z` and the observation noise covariance matrix `H` by removing rows and/or columns
        associated with the data that is not observed at this iteration. Missing values are replaced with zeros to prevent
        propagating NaNs through the computation.

        Return a binary flag tensor `all_nan_flag`,indicating if all values in the observation data are missing. This
        flag is used for numerical adjustments in the update method.

        Parameters
        ----------
        y : TensorVariable
            The observation data at time t.
        Z : TensorVariable
            The design matrix.
        H : TensorVariable
            The observation noise covariance matrix.

        Returns
        -------
        y_masked : TensorVariable
            Observation vector with missing values replaced by zeros.

        Z_masked: TensorVariable
            Design matrix adjusted to exclude the missing states from the information set of observed variables in the
            update step

        H_masked: TensorVariable
            Noise covariance matrix, adjusted to exclude the missing states

        all_nan_flag: float
            1 if the entire state vector is missing

        References
        ----------
        .. [1] Durbin, J., and S. J. Koopman. Time Series Analysis by State Space Methods.
               2nd ed, Oxford University Press, 2012.
        """
        nan_mask = pt.or_(pt.isnan(y), pt.eq(y, self.missing_fill_value))
        all_nan_flag = pt.all(nan_mask).astype(pytensor.config.floatX)
        W = pt.diag(pt.bitwise_not(nan_mask).astype(pytensor.config.floatX))

        Z_masked = W.dot(Z)
        H_masked = W.dot(H)
        y_masked = pt.set_subtensor(y[nan_mask], 0.0)

        return y_masked, Z_masked, H_masked, all_nan_flag

    @staticmethod
    def predict(a, P, c, T, R, Q) -> tuple[TensorVariable, TensorVariable]:
        """
        Perform the prediction step of the Kalman filter.

        This function computes the one-step forecast of the hidden states and the covariance matrix of the forecasted
        states, based on the current state estimates and model parameters. For computational stability, the estimated
        covariance matrix is forced to by symmetric by averaging it with its own transpose. The prediction equations
        are:

        .. math::

            \begin{align}
            a_{t+1 | t} &= T_t a_{t | t} \\
            P_{t+1 | t} &= T_t P_{t | t} T_t^T + R_t Q_t R_t^T
            \\end{align}


        Parameters
        ----------
        a : TensorVariable
            The current state vector estimate computed by the update step, a[t | t].
        P : TensorVariable
            The current covariance matrix estimate computed by the update step, P[t | t].
        c : TensorVariable
            The hidden state intercept/bias vector.
        T : TensorVariable
            The state transition matrix.
        R : TensorVariable
            The selection matrix.
        Q : TensorVariable
            The state innovation covariance matrix.

        Returns
        -------
        a_hat : TensorVariable
            One-step forecast of the hidden states
        P_hat : TensorVariable
            Covariance matrix of the forecasted hidden states

        References
        ----------
        .. [1] Durbin, J., and S. J. Koopman. Time Series Analysis by State Space Methods.
               2nd ed, Oxford University Press, 2012.
        """
        a_hat = T.dot(a) + c
        P_hat = quad_form_sym(T, P) + quad_form_sym(R, Q)

        return a_hat, P_hat

    @staticmethod
    def update(
        a, P, y, d, Z, H, all_nan_flag
    ) -> tuple[TensorVariable, TensorVariable, TensorVariable, TensorVariable, TensorVariable]:
        """
        Perform the update step of the Kalman filter.

        This function updates the state vector and covariance matrix estimates based on the current observation data,
        previous predictions, and model parameters. The filtering equations are:

        .. math::

            \begin{align}
            \\hat{y}_t &= Z_t a_{t | t-1} + d_t \\
            v_t &= y_t - \\hat{y}_t \\
            F_t &= Z_t P_{t | t-1} Z_t^T + H_t \\
            a_{t|t} &= a_{t | t-1} + P_{t | t-1} Z_t^T F_t^{-1} v_t \\
            P_{t|t} &= P_{t | t-1} - P_{t | t-1} Z_t^T F_t^{-1} Z_t P_{t | t-1}
            \\end{align}


        Parameters
        ----------
        a : TensorVariable
            The current state vector estimate, conditioned on information up to time t-1.
        P : TensorVariable
            The current covariance matrix estimate, conditioned on information up to time t-1.
        y : TensorVariable
            The observation data at time t.
        d : TensorVariable
            The matrix d.
        Z : TensorVariable
            The matrix Z.
        H : TensorVariable
            The matrix H.
        all_nan_flag : TensorVariable
            A binary flag tensor indicating whether there are any missing values in the observation data.

        Returns
        -------
        tuple[TensorVariable, TensorVariable, TensorVariable, TensorVariable, TensorVariable]
            A tuple containing the updated state vector `a_filtered`, the updated covariance matrix `P_filtered`, the
            predicted observation `obs_mu`, the predicted observation covariance matrix `obs_cov`, and the log-likelihood `ll`.
        """
        raise NotImplementedError

    def kalman_step(self, *args) -> tuple:
        """
        Performs a single iteration of the Kalman filter, which is composed of two steps : an update step and a
        prediction step. The timing convention follows [1], in which initial state and covariance estimates a0 and P0
        are taken to be predictions. As a result, the update step is applied first. The update step computes:

        .. math::

            \begin{align}
            \\hat{y}_t &= Z_t a_{t | t-1} \\
            v_t &= y_t - \\hat{y}_t \\
            F_t &= Z_t P_{t | t-1} Z_t^T + H_t \\
            a_{t|t} &= a_{t | t-1} + P_{t | t-1} Z_t^T F_t^{-1} v_t \\
            P_{t|t} &= P_{t | t-1} - P_{t | t-1} Z_t^T F_t^{-1} Z_t P_{t | t-1}
            \\end{align}

        Where the quantities :math:`a_{t|t}` and :math:`P_{t|t}` are the best linear estimates of the hidden states
        at time t, incorporating all information up to and including the observation :math:`y_t`. After the update step,
        new one-step forecasts of the hidden states can be obtained by applying the model transition dynamics in
        the prediction step:

        .. math::

            \begin{align}
            a_{t+1 | t} &= T_t a_{t | t} \\
            P_{t+1 | t} &= T_t P_{t | t} T_t^T + R_t Q_t R_t^T
            \\end{align}

        Recursive application of these two steps results in the best linear estimate of the hidden states, including
        missing values and observations subject to measurement error.

        Parameters
        ----------
        Kalman filter inputs:
            y, a, P, c, d, T, Z, R, H, Q. See the docstring for the kalman filter class for details.

        Returns
        -------
        a_filtered : TensorVariable
            Best linear estimate of hidden states given all information up to and including the present
             observation, a[t | t].

        a_hat: TensorVariable
            One-step forecast of next-period hidden states given all information up to and including the present
            observation, a[t+1 | t]

        obs_mu: TensorVariable
            Estimates of the current observation given all information available prior to the current state,
             d + Z @ a[t | t-1]

        P_filtered: TensorVariable
            Best linear estimate of the covariance between hidden states, given all information up to and including
            the present observation, P[t | t]

        P_hat: TensorVariable
            Covariance between the one-step forecasted hidden states given all information up to and including the
            present observation, P[t+1 | t]

        obs_cov: TensorVariable
            Covariance between estimated present observations, given all information available prior to the current
            state, Z @ P[t | t-1] @ Z.T + H

        ll: float
            Likelihood of the time t observation vector under the multivariate normal distribution parameterized by
            `obs_mu` and `obs_cov`

        References
        ----------
        .. [1] Durbin, J., and S. J. Koopman. Time Series Analysis by State Space Methods.
               2nd ed, Oxford University Press, 2012.
        """
        y, a, P, c, d, T, Z, R, H, Q = self.unpack_args(args)
        y_masked, Z_masked, H_masked, all_nan_flag = self.handle_missing_values(y, Z, H)

        a_filtered, P_filtered, obs_mu, obs_cov, ll = self.update(
            y=y_masked, a=a, d=d, P=P, Z=Z_masked, H=H_masked, all_nan_flag=all_nan_flag
        )

        P_filtered = stabilize(P_filtered, self.cov_jitter)

        a_hat, P_hat = self.predict(a=a_filtered, P=P_filtered, c=c, T=T, R=R, Q=Q)
        outputs = (a_filtered, a_hat, obs_mu, P_filtered, P_hat, obs_cov, ll)

        return outputs


class StandardFilter(BaseFilter):
    """
    Basic Kalman Filter
    """

    def update(self, a, P, y, d, Z, H, all_nan_flag):
        """
        Compute one-step forecasts for observed states conditioned on information up to, but not including, the current
        timestep, `y_hat`, along with the forcast covariance matrix, `F`. Marginalize over observed states to obtain
        the best linear estimate of the unobserved states, `a_filtered`, as well as the associated covariance matrix,
        `P_filtered`,  conditioned on all information, up to and including the present.

        Derivation of the Kalman filter, along with a deeper discussion of the computational elements, can be found in
        [1].

        Parameters
        ----------
        a : TensorVariable
            The current state vector estimate, conditioned on information up to time t-1.

        P : TensorVariable
            The current covariance matrix estimate, conditioned on information up to time t-1.

        y : TensorVariable
            Observations at time t.

        d : TensorVariable
            Observed state bias term.

        Z : TensorVariable
            Linear map between unobserved and observed states.

        H : TensorVariable
            Observation noise covariance matrix

        all_nan_flag : TensorVariable
            A flag indicating whether all elements in the data `y` are NaNs.

        Returns
        -------
        tuple[TensorVariable, TensorVariable, TensorVariable, TensorVariable, float]
            A tuple containing the updated state vector `a_filtered`, the updated covariance matrix `P_filtered`,
            the one-step forecast mean `y_hat`, one-step forcast covariance matrix  `F`, and the log-likelihood of
            the data, given the one-step forecasts, `ll`.

        References
        ----------
        .. [1] Durbin, J., and S. J. Koopman. Time Series Analysis by State Space Methods.
               2nd ed, Oxford University Press, 2012.
        """
        y_hat = d + Z.dot(a)
        v = y - y_hat

        PZT = P.dot(Z.T)
        F = Z.dot(PZT) + stabilize(H, self.cov_jitter)

        K = pt.linalg.solve(F.T, PZT.T, assume_a="pos", check_finite=False).T
        I_KZ = pt.eye(self.n_states) - K.dot(Z)

        a_filtered = a + K.dot(v)
        P_filtered = quad_form_sym(I_KZ, P) + quad_form_sym(K, H)

        F_inv_v = pt.linalg.solve(F, v, assume_a="pos", check_finite=False)
        inner_term = v.T @ F_inv_v

        F_logdet = pt.log(pt.linalg.det(F))

        ll = pt.switch(
            all_nan_flag,
            0.0,
            -0.5 * (MVN_CONST + F_logdet + inner_term).ravel()[0],
        )

        return a_filtered, P_filtered, y_hat, F, ll


class SquareRootFilter(BaseFilter):
    """
    Kalman filter with Cholesky factorization

    Kalman filter implementation using a Cholesky factorization plus pt.solve_triangular to (attempt) to speed up
    inversion of the observation covariance matrix `F`.

    """

    def predict(self, a, P, c, T, R, Q):
        """
        Compute one-step forecasts for the hidden states conditioned on information up to, but not including, the current
        timestep, `a_hat`, along with the forcast covariance matrix, `P_hat`.

        .. warning::
            Very important -- In this function, $P$ is the **cholesky factor** of the covariance matrix, not the
            covariance matrix itself. The name `P` is kept for consistency with the superclass.
        """
        # Rename P to P_chol for clarity
        P_chol = P

        a_hat = T.dot(a) + c
        Q_chol = pt.linalg.cholesky(Q, lower=True)

        M = pt.horizontal_stack(T @ P_chol, R @ Q_chol).T
        R_decomp = pt.linalg.qr(M, mode="r")
        P_chol_hat = R_decomp[: self.n_states, : self.n_states].T

        return a_hat, P_chol_hat

    def update(self, a, P, y, d, Z, H, all_nan_flag):
        """
        Compute posterior estimates of the hidden state distributions conditioned on the observed data, up to and
        including the present timestep. Also compute the log-likelihood of the data given the one-step forecasts.

        .. warning::
            Very important -- In this function, $P$ is the **cholesky factor** of the covariance matrix, not the
            covariance matrix itself. The name `P` is kept for consistency with the superclass.
        """

        # Rename P to P_chol for clarity
        P_chol = P

        y_hat = Z.dot(a) + d
        v = y - y_hat

        H_chol = pytensor.ifelse(pt.all(pt.eq(H, 0.0)), H, pt.linalg.cholesky(H, lower=True))

        # The following notation comes from https://ipnpr.jpl.nasa.gov/progress_report/42-233/42-233A.pdf
        # Construct upper-triangular block matrix A = [[chol(H), Z @ L_pred],
        #                                              [0,           L_pred]]
        # The Schur decomposition of this matrix will be B (upper triangular). We are
        # more insterested in B^T:
        # Structure of B^T = [[chol(F),     0              ],
        #                    [K @ chol(F), chol(P_filtered)]
        zeros = pt.zeros((self.n_states, self.n_endog))
        upper = pt.horizontal_stack(H_chol, Z @ P_chol)
        lower = pt.horizontal_stack(zeros, P_chol)
        A_T = pt.vertical_stack(upper, lower)
        B = pt.linalg.qr(A_T.T, mode="r").T

        F_chol = B[: self.n_endog, : self.n_endog]
        K_F_chol = B[self.n_endog :, : self.n_endog]
        P_chol_filtered = B[self.n_endog :, self.n_endog :]

        def compute_non_degenerate(P_chol_filtered, F_chol, K_F_chol, v):
            a_filtered = a + K_F_chol @ solve_triangular(F_chol, v, lower=True)

            inner_term = solve_triangular(
                F_chol, solve_triangular(F_chol, v, lower=True), lower=True
            )
            loss = (v.T @ inner_term).ravel()

            # abs necessary because we're not guaranteed a positive diagonal from the schur decomposition
            logdet = 2 * pt.log(pt.abs(pt.diag(F_chol))).sum()

            ll = -0.5 * (self.n_endog * (MVN_CONST + logdet) + loss)[0]

            return [a_filtered, P_chol_filtered, ll]

        def compute_degenerate(P_chol_filtered, F_chol, K_F_chol, v):
            """
            If F is zero (usually because there were no observations this period), then we want:
            K = 0, a = a, P = P, ll = 0
            """
            return [a, P_chol, pt.zeros(())]

        [a_filtered, P_chol_filtered, ll] = pytensor.ifelse(
            pt.eq(all_nan_flag, 1.0),
            compute_degenerate(P_chol_filtered, F_chol, K_F_chol, v),
            compute_non_degenerate(P_chol_filtered, F_chol, K_F_chol, v),
        )

        a_filtered = pt.specify_shape(a_filtered, (self.n_states,))
        P_chol_filtered = pt.specify_shape(P_chol_filtered, (self.n_states, self.n_states))

        return a_filtered, P_chol_filtered, y_hat, F_chol, ll

    def _postprocess_scan_results(self, results, a0, P0, n) -> list[TensorVariable]:
        """
        Convert the Cholesky factor of the covariance matrix back to the covariance matrix itself.
        """
        results = super()._postprocess_scan_results(results, a0, P0, n)
        (
            filtered_states,
            predicted_states,
            observed_states,
            filtered_covariances_cholesky,
            predicted_covariances_cholesky,
            observed_covariances_cholesky,
            loglike_obs,
        ) = results

        def square_sequnece(L, k):
            X = pt.einsum("...ij,...kj->...ik", L, L.copy())
            X = pt.specify_shape(X, (n, k, k))
            return X

        filtered_covariances = square_sequnece(filtered_covariances_cholesky, k=self.n_states)
        predicted_covariances = square_sequnece(predicted_covariances_cholesky, k=self.n_states)
        observed_covariances = square_sequnece(observed_covariances_cholesky, k=self.n_endog)

        return [
            filtered_states,
            predicted_states,
            observed_states,
            filtered_covariances,
            predicted_covariances,
            observed_covariances,
            loglike_obs,
        ]


class UnivariateFilter(BaseFilter):
    """
    The univariate kalman filter, described in [1], section 6.4.2, avoids inversion of the F matrix, as well as two
    matrix multiplications, at the cost of an additional loop. Note that the name doesn't mean there's only one
    observed time series. This is called univariate because it updates the state mean and covariance matrices one
    variable at a time, using an inner-inner loop.

    This is useful when states are perfectly observed, because the F matrix can easily become degenerate in these cases.

    References
    ----------
    .. [1] Durbin, J., and S. J. Koopman. Time Series Analysis by State Space Methods.
            2nd ed, Oxford University Press, 2012.

    """

    def _univariate_inner_filter_step(self, y, Z_row, d_row, sigma_H, nan_flag, a, P):
        y_hat = Z_row.dot(a) + d_row
        v = y - y_hat

        PZT = P.dot(Z_row.T)
        F = Z_row.dot(PZT) + sigma_H

        # Set the zero flag for F first, then jitter it to avoid a divide-by-zero NaN later
        F_zero_flag = pt.or_(pt.eq(F, 0), nan_flag)
        F = F + self.cov_jitter

        # If F is zero (implies y is NAN or another degenerate case), then we want:
        # K = 0, a = a, P = P, ll = 0
        K = PZT / F * (1 - F_zero_flag)

        a_filtered = a + K * v
        P_filtered = P - pt.outer(K, K) * F

        ll_inner = pt.switch(F_zero_flag, 0.0, pt.log(F) + v**2 / F)

        return a_filtered, P_filtered, pt.atleast_1d(y_hat), pt.atleast_2d(F), ll_inner

    def kalman_step(self, y, a, P, c, d, T, Z, R, H, Q):
        nan_mask = pt.isnan(y)

        W = pt.set_subtensor(pt.eye(y.shape[0])[nan_mask, nan_mask], 0.0)
        Z_masked = W.dot(Z)
        H_masked = W.dot(H)
        y_masked = pt.set_subtensor(y[nan_mask], 0.0)

        result, updates = pytensor.scan(
            self._univariate_inner_filter_step,
            sequences=[y_masked, Z_masked, d, pt.diag(H_masked), nan_mask],
            outputs_info=[a, P, None, None, None],
            name="univariate_inner_scan",
        )

        a_filtered, P_filtered, obs_mu, obs_cov, ll_inner = result
        a_filtered, P_filtered, obs_mu, obs_cov = (
            a_filtered[-1],
            P_filtered[-1],
            obs_mu[-1],
            obs_cov[-1],
        )

        P_filtered = stabilize(0.5 * (P_filtered + P_filtered.T), self.cov_jitter)
        a_hat, P_hat = self.predict(a=a_filtered, P=P_filtered, c=c, T=T, R=R, Q=Q)

        ll = -0.5 * ((pt.neq(ll_inner, 0).sum()) * MVN_CONST + ll_inner.sum())

        return a_filtered, a_hat, obs_mu, P_filtered, P_hat, obs_cov, ll



from collections.abc import Sequence
from typing import Any

import pytensor
import pytensor.tensor as pt

from pymc_extras.statespace.core.statespace import PyMCStateSpace
from pymc_extras.statespace.models.utilities import make_default_coords
from pymc_extras.statespace.utils.constants import (
    ALL_STATE_AUX_DIM,
    ALL_STATE_DIM,
    AR_PARAM_DIM,
    ERROR_AR_PARAM_DIM,
    EXOG_STATE_DIM,
    FACTOR_DIM,
    OBS_STATE_AUX_DIM,
    OBS_STATE_DIM,
    TIME_DIM,
)

floatX = pytensor.config.floatX


class BayesianDynamicFactor(PyMCStateSpace):
    r"""
    Dynamic Factor Models

    Notes
    -----
    The Dynamic Factor Model (DFM) is a multivariate state-space model used to represent high-dimensional time series
    as being driven by a smaller set of unobserved dynamic factors.

    Given a set of observed time series :math:`\{y_t\}_{t=0}^T`, where

    .. math::
        y_t = \begin{bmatrix} y_{1,t} & y_{2,t} & \cdots & y_{k_{\text{endog}},t} \end{bmatrix}^T,

    the DFM assumes that each series is a linear combination of a few latent factors and (optionally) autoregressive errors.

    Let:
    - :math:`k` be the number of dynamic factors (k_factors),
    - :math:`p` be the order of the latent factor process (factor_order),
    - :math:`q` be the order of the observation error process (error_order).

    The model equations are in reduced form is:

    .. math::
        y_t &= \Lambda f_t + B x_t + u_t + \eta_t \\
        f_t &= A_1 f_{t-1} + \cdots + A_p f_{t-p} + \varepsilon_{f,t} \\
        u_t &= C_1 u_{t-1} + \cdots + C_q u_{t-q} + \varepsilon_{u,t}

    Where:
    - :math:`f_t` is the vector of latent dynamic factors (size :math:`k`),
    - :math:`x_t` is an optional vector of exogenous variables
    - :math:`u_t` is a vector of autoregressive observation errors (if `error_var=True` with a VAR(q) structure, else treated as independent AR processes),
    - :math:`\eta_t \sim \mathcal{N}(0, H_t)` is an optional measurement error (if `measurement_error=True`),
    - :math:`\varepsilon_{f,t} \sim \mathcal{N}(0, I)` and :math:`\varepsilon_{u,t} \sim \mathcal{N}(0, \Sigma_u)` are independent noise terms.
        To identify the factors, the innovations to the factor process are standardized with identity covariance.

    Internally, the model is represented in state-space form by stacking all current and lagged latent factors and (if present)
    AR observation errors into a single state vector of dimension:  :math:: k_{\text{states}} = k \cdot p + k_{\text{endog}} \cdot q,
    where :math:`k_{\text{endog}}` is the number of observed time series.

    The state vector is defined as:

    .. math::
        s_t = \begin{bmatrix}
            f_t(1) \\
            \vdots \\
            f_t(k) \\
            f_{t-p+1}(1) \\
            \vdots \\
            f_{t-p+1}(k) \\
            u_t(1) \\
            \vdots \\
            u_t(k_{\text{endog}}) \\
            \vdots \\
            u_{t-q+1}(1) \\
            \vdots \\
            u_{t-q+1}(k_{\text{endog}})
        \end{bmatrix}
        \in \mathbb{R}^{k_{\text{states}}}

    The transition equation is given by:

    .. math::
        s_{t+1} = T s_t + R \epsilon_t

    Where:
    - :math:`T` is the state transition matrix, composed of:
        - VAR coefficients :math:`A_1, \dots, A_{p*k_factors}` for the factors,
        - (if enabled) AR coefficients :math:`C_1, \dots, C_q` for the observation errors.
        .. math::
            T = \begin{bmatrix}
            A_{1,1}  &   A_{1,2}  &   \cdots  &   A_{1,p}  &   0       &   0       &   \cdots  &   0 \\
            A_{2,1}  &   A_{2,2}  &   \cdots  &   A_{2,p}  &   0       &   0       &   \cdots  &   0 \\
            1       &   0       &   \cdots  &   0          &   0       &   0       &   \cdots  &   0 \\
            0       &   1       &   \cdots  &   0          &   0       &   0       &   \cdots  &   0 \\
            \vdots  &   \vdots  &   \ddots  &   \vdots     &   \vdots  &   \vdots  &   \ddots  &   \vdots \\
            \hline
            0       &   0       &   \cdots  &   0       &   C_{1,1}  &  \cdots  &    C_{1,2} &   C_{1,q} \\
            0       &   0       &   \cdots  &   0       &   1       &   0       &   \cdots  &   0 \\
            0       &   0       &   \cdots  &   0       &   0       &   1       &   \cdots  &   0 \\
            \vdots  &   \vdots  &           &   \vdots  &   \vdots  &   \vdots  &   \ddots  &   \vdots
            \end{bmatrix}
            \in \mathbb{R}^{k_{\text{states}} \times k_{\text{states}}}

    - :math:`\epsilon_t` contains the independent shocks (innovations) and has dimension :math:`k + k_{\text{endog}}` if AR errors are included.
        .. math::
            \epsilon_t = \begin{bmatrix}
            \epsilon_{f,t} \\
            \epsilon_{u,t}
            \end{bmatrix}
            \in \mathbb{R}^{k +  k_{\text{endog}}}

    - :math:`R` is a selection matrix mapping shocks to state transitions.
        .. math::
            R = \begin{bmatrix}
            1       &   0       &   \cdots  &   0       &   0       &   0       &   \cdots  &   0 \\
            0       &   1       &   \cdots  &   0       &   0       &   0       &   \cdots  &   0 \\
            \vdots  &   \vdots  &   \ddots  &   \vdots  &   \vdots  &   \vdots  &   \ddots  &   \vdots \\
            0       &   0       &   \cdots  &   1       &   0       &   0       &   \cdots  &   0 \\
            0       &   0       &   \cdots  &   0       &   1       &   0       &   \cdots  &   0 \\
            0       &   0       &   \cdots  &   0       &   0       &   1       &   \cdots  &   0 \\
            \vdots  &   \vdots  &   \ddots  &   \vdots  &   \vdots  &   \vdots  &   \ddots  &   \vdots \\
            \end{bmatrix}
            \in \mathbb{R}^{k_{\text{states}} \times (k + k_{\text{endog}})}

    The observation equation is given by:

    .. math::

        y_t = Z s_t + \eta_t

    where

    - :math:`y_t` is the vector of observed variables at time :math:`t`

    - :math:`Z` is the design matrix of the state space representation
        .. math::
            Z = \begin{bmatrix}
            \lambda_{1,1}       &   \lambda_{1,k}   &   \vdots    &   1   &   0   &   \cdots  &   0   &   0   &   \cdots  &   0 \\
            \lambda_{2,1}       &   \lambda_{2,k}   &   \vdots    &   0   &   1   &   \cdots   &   0   &   \cdots  &   0 \\
            \vdots              &   \vdots          &   \vdots  &   \vdots  &   \ddots  &   \vdots  &   \vdots  &   \ddots  &   \vdots \\
            \lambda_{k_{\text{endog}},1}  &   \cdots  &   \lambda_{k_{\text{endog}},k}  &   0   &   0   &   \cdots  &   1   &   0   &   \cdots  &   0 \\
            \end{bmatrix}
            \in \mathbb{R}^{k_{\text{endog}} \times k_{\text{states}}}

    - :math:`\eta_t` is the vector of observation errors at time :math:`t`

    When exogenous variables :math:`x_t` are present, the implementation follows `pymc_extras/statespace/models/structural/components/regression.py`.
    In this case, the state vector is extended to include the beta parameters, and the design matrix is modified accordingly,
    becoming 3-dimensional to handle time-varying exogenous regressors.
    This approach provides greater flexibility, controlled by the boolean flags `shared_exog_state` and `exog_innovations`.
    Unlike Statsmodels, where exogenous variables are included only in the observation equation, here they are fully integrated into the state-space
    representation.

    .. warning::

        Identification can be an issue, particularly when many observed series load onto only a few latent factors.
        These models are only identified up to a sign flip in the factor loadings. Proper prior specification is crucial
        for good estimation and inference.

    Examples
    --------
    The following code snippet estimates a dynamic factor model with 1 latent factors,
    a AR(2) structure on the factor and a AR(1) structure on the errors:

    .. code:: python

        import pymc_extras.statespace as pmss
        import pymc as pm

        # Create DFM Statespace Model
        dfm_mod = pmss.BayesianDynamicFactor(
                k_factors=1,                # Number of latent dynamic factors
                factor_order=2,             # Number of lags for the latent factor process
                endog_names=data.columns,   # Names of the observed time series (endogenous variables) (we could also use k_endog = len(data.columns))
                error_order=1,              # Order of the autoregressive process for the observation noise (i.e., AR(q) error, here q=1)
                error_var=False,            # If False, models errors as separate AR processes
                error_cov_type="diagonal",  # Structure of the observation error covariance matrix: uncorrelated noise across series
                measurement_error=True,     # Whether to include a measurement error term in the model
                verbose=True
            )

        # Unpack coords
        coords = dfm_mod.coords


        with pm.Model(coords=coords) as pymc_mod:
            # Priors for the initial state mean and covariance
            x0 = pm.Normal("x0", dims=["state_dim"])
            P0 = pm.HalfNormal("P0", dims=["state_dim", "state_dim"])

            # Factor loadings: shape (k_endog, k_factors)
            factor_loadings = pm.Normal("factor_loadings", sigma=1, dims=["k_endog", "k_factors"])

            # AR coefficients for factor dynamics: shape (k_factors, factor_order)
            factor_ar = pm.Normal("factor_ar", sigma=1, dims=["k_factors", "k_factors" * "factor_order"])

            # AR coefficients for observation noise: shape (k_endog, error_order)
            error_ar = pm.Normal("error_ar", sigma=1, dims=["k_endog", "error_order"])

            # Std devs for observation noise: shape (k_endog,)
            error_sigma = pm.HalfNormal("error_sigma", dims=["k_endog"])

            # Observation noise covariance matrix
            obs_sigma = pm.HalfNormal("sigma_obs", dims=["k_endog"])

            # Build the symbolic graph and attach it to the model
            dfm_mod.build_statespace_graph(data=data, mode="JAX")

            # Sampling
            idata = pm.sample(
                draws=500,
                chains=2,
                nuts_sampler="nutpie",
                nuts_sampler_kwargs={"backend": "jax", "gradient_backend": "jax"},
            )

    """

    def __init__(
        self,
        k_factors: int,
        factor_order: int,
        k_endog: int | None = None,
        endog_names: Sequence[str] | None = None,
        k_exog: int | None = None,
        exog_names: Sequence[str] | None = None,
        shared_exog_states: bool = False,
        exog_innovations: bool = False,
        error_order: int = 0,
        error_var: bool = False,
        error_cov_type: str = "diagonal",
        measurement_error: bool = False,
        verbose: bool = True,
    ):
        """
        Create a Bayesian Dynamic Factor Model.

        Parameters
        ----------
        k_factors : int
            Number of latent factors.

        factor_order : int
            Order of the VAR process for the latent factors. If set to 0, the factors have no autoregressive dynamics
            and are modeled as a white noise process, i.e., :math:`f_t = \varepsilon_{f,t}`.
            Therefore, the state vector will include one state per factor and "factor_ar" will not exist.

        k_endog : int, optional
            Number of observed time series. If not provided, the number of observed series will be inferred from `endog_names`.
            At least one of `k_endog` or `endog_names` must be provided.

        endog_names : list of str, optional
            Names of the observed time series. If not provided, default names will be generated as `endog_1`, `endog_2`, ..., `endog_k` based on `k_endog`.
            At least one of `k_endog` or `endog_names` must be provided.

        k_exog : int, optional
            Number of exogenous variables. If not provided, the model will not have exogenous variables.

        exog_names : Sequence[str], optional
            Names of the exogenous variables. If not provided, but `k_exog` is specified, default names will be generated as `exog_1`, `exog_2`, ..., `exog_k`.

        shared_exog_states: bool, optional
            Whether exogenous latent states are shared across the observed states. If True, there will be only one set of exogenous latent
            states, which are observed by all observed states. If False, each observed state has its own set of exogenous latent states.

        exog_innovations : bool, optional
            Whether to allow time-varying regression coefficients. If True, coefficients follow a random walk.

        error_order : int, optional
            Order of the AR process for the observation error component.
            Default is 0, corresponding to white noise errors.

        error_var : bool, optional
            If True, errors are modeled jointly via a VAR process;
            otherwise, each error is modeled separately.

        error_cov_type : {'scalar', 'diagonal', 'unstructured'}, optional
            Structure of the covariance matrix of the observation errors.

        measurement_error: bool, default True
            If true, a measurement error term is added to the model.

        verbose: bool, default True
            If true, a message will be logged to the terminal explaining the variable names, dimensions, and supports.

        """

        if k_endog is None and endog_names is None:
            raise ValueError("Either k_endog or endog_names must be provided.")
        if k_endog is None:
            k_endog = len(endog_names)
        if endog_names is None:
            endog_names = [f"endog_{i}" for i in range(k_endog)]

        self.endog_names = endog_names
        self.k_endog = k_endog
        self.k_factors = k_factors
        self.factor_order = factor_order
        self.error_order = error_order
        self.error_var = error_var
        self.error_cov_type = error_cov_type

        if k_exog is None and exog_names is None:
            self.k_exog = 0
        else:
            self.shared_exog_states = shared_exog_states
            self.exog_innovations = exog_innovations
            if k_exog is None:
                k_exog = len(exog_names) if exog_names is not None else 0
            elif exog_names is None:
                exog_names = [f"exog_{i}" for i in range(k_exog)] if k_exog > 0 else None
            self.k_exog = k_exog
            self.exog_names = exog_names

        self.k_exog_states = self.k_exog * self.k_endog if not shared_exog_states else self.k_exog
        self.exog_flag = self.k_exog > 0

        # Determine the dimension for the latent factor states.
        # For static factors, one use k_factors.
        # For dynamic factors with lags, the state include current factors and past lags.
        # If factor_order is 0, we treat the factor as static (no dynamics),
        # but it is still included in the state vector with one state per factor. Factor_ar paramter will not exist in this case.
        k_factor_states = max(self.factor_order, 1) * k_factors

        # Determine the dimension for the error component.
        # If error_order > 0 then we add additional states for error dynamics, otherwise white noise error.
        k_error_states = k_endog * error_order if error_order > 0 else 0

        # Total state dimension
        k_states = k_factor_states + k_error_states + self.k_exog_states

        # Number of independent shocks.
        # Typically, the latent factors introduce k_factors shocks.
        # If error_order > 0 and errors are modeled jointly or separately, add appropriate count.
        k_posdef = k_factors + (k_endog if error_order > 0 else 0) + self.k_exog_states
        # k_posdef = (k_factors + (k_endog if error_order > 0 else 0) + self.k_exog_states if self.exog_innovations else 0)

        # Initialize the PyMCStateSpace base class.
        super().__init__(
            k_endog=k_endog,
            k_states=k_states,
            k_posdef=k_posdef,
            verbose=verbose,
            measurement_error=measurement_error,
        )

    @property
    def param_names(self):
        names = [
            "x0",
            "P0",
            "factor_loadings",
            "factor_ar",
            "error_ar",
            "error_sigma",
            "error_cov",
            "sigma_obs",
            "beta",
            "beta_sigma",
        ]

        # Handle cases where parameters should be excluded based on model settings
        if self.factor_order == 0:
            names.remove("factor_ar")
        if self.error_order == 0:
            names.remove("error_ar")
        if self.error_cov_type in ["scalar", "diagonal"]:
            names.remove("error_cov")
        if self.error_cov_type == "unstructured":
            names.remove("error_sigma")
        if not self.measurement_error:
            names.remove("sigma_obs")
        if not self.exog_flag:
            names.remove("beta")
            names.remove("beta_sigma")
        if self.exog_flag and not self.exog_innovations:
            names.remove("beta_sigma")

        return names

    @property
    def param_info(self) -> dict[str, dict[str, Any]]:
        info = {
            "x0": {
                "shape": (self.k_states,),
                "constraints": None,
            },
            "P0": {
                "shape": (self.k_states, self.k_states),
                "constraints": "Positive Semi-definite",
            },
            "factor_loadings": {
                "shape": (self.k_endog, self.k_factors),
                "constraints": None,
            },
            "factor_ar": {
                "shape": (self.k_factors, self.factor_order * self.k_factors),
                "constraints": None,
            },
            "error_ar": {
                "shape": (
                    self.k_endog,
                    self.error_order * self.k_endog if self.error_var else self.error_order,
                ),
                "constraints": None,
            },
            "error_sigma": {
                "shape": (self.k_endog,) if self.error_cov_type == "diagonal" else (),
                "constraints": "Positive",
            },
            "error_cov": {
                "shape": (self.k_endog, self.k_endog),
                "constraints": "Positive Semi-definite",
            },
            "sigma_obs": {
                "shape": (self.k_endog,),
                "constraints": "Positive",
            },
            "beta": {
                "shape": (self.k_exog_states,),
                "constraints": None,
            },
            "beta_sigma": {
                "shape": (self.k_exog_states,),
                "constraints": "Positive",
            },
        }

        for name in self.param_names:
            info[name]["dims"] = self.param_dims[name]

        return {name: info[name] for name in self.param_names}

    @property
    def state_names(self) -> list[str]:
        """
        Returns the names of the hidden states: first factor states (with lags),
        idiosyncratic error states (with lags), then exogenous states.
        """
        names = [
            f"L{lag}.factor_{i}"
            for i in range(self.k_factors)
            for lag in range(max(self.factor_order, 1))
        ]

        if self.error_order > 0:
            names.extend(
                f"L{lag}.error_{i}" for i in range(self.k_endog) for lag in range(self.error_order)
            )

        if self.exog_flag:
            if self.shared_exog_states:
                names.extend([f"beta_{exog_name}[shared]" for exog_name in self.exog_names])
            else:
                names.extend(
                    f"beta_{exog_name}[{endog_name}]"
                    for exog_name in self.exog_names
                    for endog_name in self.endog_names
                )
        return names

    @property
    def observed_states(self) -> list[str]:
        """
        Returns the names of the observed states (i.e., the endogenous variables).
        """
        return self.endog_names

    @property
    def coords(self) -> dict[str, Sequence]:
        coords = make_default_coords(self)

        coords[FACTOR_DIM] = [f"factor_{i+1}" for i in range(self.k_factors)]

        if self.factor_order > 0:
            coords[AR_PARAM_DIM] = list(range(1, (self.factor_order * self.k_factors) + 1))

        if self.error_order > 0:
            if self.error_var:
                coords[ERROR_AR_PARAM_DIM] = list(range(1, (self.error_order * self.k_endog) + 1))
            else:
                coords[ERROR_AR_PARAM_DIM] = list(range(1, self.error_order + 1))

        if self.exog_flag:
            coords[EXOG_STATE_DIM] = list(range(1, self.k_exog_states + 1))

        return coords

    @property
    def shock_names(self) -> list[str]:
        shock_names = [f"factor_shock_{i}" for i in range(self.k_factors)]

        if self.error_order > 0:
            shock_names.extend(f"error_shock_{i}" for i in range(self.k_endog))

        if self.exog_flag:
            if self.shared_exog_states:
                shock_names.extend(f"exog_shock_{i}.shared" for i in range(self.k_exog))
            else:
                shock_names.extend(
                    f"exog_shock_{i}.endog_{j}"
                    for i in range(self.k_exog)
                    for j in range(self.k_endog)
                )

        return shock_names

    @property
    def param_dims(self):
        coord_map = {
            "x0": (ALL_STATE_DIM,),
            "P0": (ALL_STATE_DIM, ALL_STATE_AUX_DIM),
            "factor_loadings": (OBS_STATE_DIM, FACTOR_DIM),
        }
        if self.factor_order > 0:
            coord_map["factor_ar"] = (FACTOR_DIM, AR_PARAM_DIM)

        if self.error_order > 0:
            coord_map["error_ar"] = (OBS_STATE_DIM, ERROR_AR_PARAM_DIM)

        if self.error_cov_type in ["scalar"]:
            coord_map["error_sigma"] = ()

        elif self.error_cov_type in ["diagonal"]:
            coord_map["error_sigma"] = (OBS_STATE_DIM,)

        if self.error_cov_type == "unstructured":
            coord_map["error_cov"] = (OBS_STATE_DIM, OBS_STATE_AUX_DIM)

        if self.measurement_error:
            coord_map["sigma_obs"] = (OBS_STATE_DIM,)

        if self.exog_flag:
            coord_map["beta"] = (EXOG_STATE_DIM,)
            if self.exog_innovations:
                coord_map["beta_sigma"] = (EXOG_STATE_DIM,)

        return coord_map

    @property
    def data_info(self):
        if self.exog_flag:
            return {
                "exog_data": {
                    "shape": (None, self.k_exog),
                    "dims": (TIME_DIM, EXOG_STATE_DIM),
                },
            }
        return {}

    @property
    def data_names(self):
        if self.exog_flag:
            return ["exog_data"]
        return []

    def make_symbolic_graph(self):
        if not self.exog_flag:
            x0 = self.make_and_register_variable("x0", shape=(self.k_states,), dtype=floatX)
        else:
            initial_factor_loadings = self.make_and_register_variable(
                "x0", shape=(self.k_states - self.k_exog_states,), dtype=floatX
            )
            initial_betas = self.make_and_register_variable(
                "beta", shape=(self.k_exog_states,), dtype=floatX
            )
            x0 = pt.concatenate([initial_factor_loadings, initial_betas], axis=0)

        self.ssm["initial_state", :] = x0

        # Initial covariance
        P0 = self.make_and_register_variable(
            "P0", shape=(self.k_states, self.k_states), dtype=floatX
        )
        self.ssm["initial_state_cov", :, :] = P0

        # Design matrix (Z)
        # Construction with block structure:
        # When factor_order <= 1 and error_order = 0:
        #   [ A ]               A is the factor loadings matrix with shape (k_endog, k_factors)
        #
        # When factor_order > 1, add block of zeros for the factors lags:
        #   [ A | 0 ]           the zero block has shape (k_endog, k_factors * (factor_order - 1))
        #
        # When error_order > 0, add identity matrix and additional zero block for errors lags:
        #   [ A | 0 | I | 0 ]   I is the identity matrix (k_endog, k_endog) and the final zero block
        #                       has shape (k_endog, k_endog * (error_order - 1))
        #
        # When exog_flag=True, exogenous data (exog_data) is included and the design
        # matrix becomes 3D with the first dimension indexing time:
        #   - shared_exog_states=True: exog_data is broadcast across all endogenous series
        #       → shape (n_timepoints, k_endog, k_exog)
        #   - shared_exog_states=False: each endogenous series gets its own exog block
        #       → block-diagonal structure with shape (n_timepoints, k_endog, k_exog * k_endog)
        # In this case, the base design matrix (factors + errors) is repeated over
        # time and concatenated with the exogenous block. The final design matrix
        # has shape (n_timepoints, k_endog, n_columns) and combines all components.
        factor_loadings = self.make_and_register_variable(
            "factor_loadings", shape=(self.k_endog, self.k_factors), dtype=floatX
        )
        # Add factor loadings (A matrix)
        matrix_parts = [factor_loadings]

        # Add zero block for the factors lags when factor_order > 1
        if self.factor_order > 1:
            matrix_parts.append(
                pt.zeros((self.k_endog, self.k_factors * (self.factor_order - 1)), dtype=floatX)
            )
        # Add identity and zero blocks for error lags when error_order > 0
        if self.error_order > 0:
            error_matrix = pt.eye(self.k_endog, dtype=floatX)
            matrix_parts.append(error_matrix)
            matrix_parts.append(
                pt.zeros((self.k_endog, self.k_endog * (self.error_order - 1)), dtype=floatX)
            )
        if len(matrix_parts) == 1:
            design_matrix = factor_loadings * 1.0  # copy to ensure a new PyTensor variable
            design_matrix.name = "design"
            # TODO: This is a hack to ensure the design matrix isn't identically equal to the factor_loadings when error_order=0 and factor_order=0
        else:
            design_matrix = pt.concatenate(matrix_parts, axis=1)
            design_matrix.name = "design"
        # Handle exogenous variables (if any)
        if self.exog_flag:
            exog_data = self.make_and_register_data("exog_data", shape=(None, self.k_exog))
            if self.shared_exog_states:
                # Shared exogenous states: same exog data is used across all endogenous variables
                # Shape becomes (n_timepoints, k_endog, k_exog)
                Z_exog = pt.specify_shape(
                    pt.join(1, *[pt.expand_dims(exog_data, 1) for _ in range(self.k_endog)]),
                    (None, self.k_endog, self.k_exog),
                )
            else:
                # Separate exogenous states: each endogenous variable gets its own exog block
                # Create block-diagonal structure and reshape to (n_timepoints, k_endog, k_exog * k_endog)
                Z_exog = pt.linalg.block_diag(
                    *[pt.expand_dims(exog_data, 1) for _ in range(self.k_endog)]
                )
                Z_exog = pt.specify_shape(Z_exog, (None, self.k_endog, self.k_exog * self.k_endog))

            # Repeat base design_matrix over time dimension to match exogenous time series
            n_timepoints = Z_exog.shape[0]
            design_matrix_time = pt.tile(design_matrix, (n_timepoints, 1, 1))
            # Concatenate the repeated design matrix with exogenous matrix along the last axis
            # Final shape: (n_timepoints, k_endog, n_columns + n_exog_columns)
            design_matrix = pt.concatenate([design_matrix_time, Z_exog], axis=2)

        self.ssm["design"] = design_matrix

        # Transition matrix (T)
        # Construction with block-diagonal structure:
        # Each latent component (factors, errors, exogenous states) contributes its own transition block,
        # and the full transition matrix is assembled with block_diag.
        #   T = block_diag(A, B, C)
        #
        # - Factors (block A):
        #   If factor_order > 0, the factor AR coefficients are organized into a
        #   VAR(p) companion matrix of size (k_factors * factor_order, k_factors * factor_order).
        #   This block shifts lagged factor states and applies AR coefficients.
        #   If factor_order = 0, a zero matrix is used instead.
        #
        # - Errors (block B):
        #   If error_order > 0:
        #     * error_var=True → build a full VAR(p) companion matrix (cross-series correlations allowed).
        #     * error_var=False → build independent AR(p) companion matrices (no cross-series effects).
        #
        # - Exogenous states (block C):
        #   If exog_flag=True, exogenous states are either constant or follow a random walk, modeled with an identity
        #       transition block of size (k_exog_states, k_exog_states).
        #
        # The final transition matrix is block-diagonal, combining all active components:
        #   Transition = block_diag(Factors, Errors, Exogenous)

        # auxiliary functions to build transition matrix block
        def build_var_block_matrix(ar_coeffs, k_series, p):
            """
            Build the VAR(p) companion matrix for the factors.

            ar_coeffs: PyTensor matrix of shape (k_series, p * k_series)
                    [A1 | A2 | ... | Ap] horizontally concatenated.
            k_series: number of series
            p: lag order
            """
            size = k_series * p
            block = pt.zeros((size, size), dtype=floatX)

            # First block row: the AR coefficient matrices for each lag
            block = block[0:k_series, 0 : k_series * p].set(ar_coeffs)

            # Sub-diagonal identity blocks (shift structure)
            if p > 1:
                # Create the identity pattern for all sub-diagonal blocks
                identity_pattern = pt.eye(k_series * (p - 1), dtype=floatX)
                block = block[k_series:, : k_series * (p - 1)].set(identity_pattern)

            return block

        def build_independent_var_block_matrix(ar_coeffs, k_series, p):
            """
            Build a VAR(p)-style companion matrix for independent AR(p) processes
            with interleaved state ordering:
            (x1(t), x2(t), ..., x1(t-1), x2(t-1), ...).

            ar_coeffs: PyTensor matrix of shape (k_series, p)
            k_series: number of independent series
            p: lag order
            """
            size = k_series * p
            block = pt.zeros((size, size), dtype=floatX)

            # First block row: AR coefficients per series (block diagonal)
            for j in range(k_series):
                for lag in range(p):
                    col_idx = lag * k_series + j
                    block = pt.set_subtensor(block[j, col_idx], ar_coeffs[j, lag])

            # Sub-diagonal identity blocks (shift)
            if p > 1:
                identity_pattern = pt.eye(k_series * (p - 1), dtype=floatX)
                block = pt.set_subtensor(block[k_series:, : k_series * (p - 1)], identity_pattern)
            return block

        transition_blocks = []
        # Block A: Factors
        if self.factor_order > 0:
            factor_ar = self.make_and_register_variable(
                "factor_ar",
                shape=(self.k_factors, self.factor_order * self.k_factors),
                dtype=floatX,
            )
            transition_blocks.append(
                build_var_block_matrix(factor_ar, self.k_factors, self.factor_order)
            )
        else:
            transition_blocks.append(pt.zeros((self.k_factors, self.k_factors), dtype=floatX))
        # Block B: Errors
        if self.error_order > 0 and self.error_var:
            error_ar = self.make_and_register_variable(
                "error_ar", shape=(self.k_endog, self.error_order * self.k_endog), dtype=floatX
            )
            transition_blocks.append(
                build_var_block_matrix(error_ar, self.k_endog, self.error_order)
            )
        elif self.error_order > 0 and not self.error_var:
            error_ar = self.make_and_register_variable(
                "error_ar", shape=(self.k_endog, self.error_order), dtype=floatX
            )
            transition_blocks.append(
                build_independent_var_block_matrix(error_ar, self.k_endog, self.error_order)
            )
        # Block C: Exogenous states
        if self.exog_flag:
            transition_blocks.append(pt.eye(self.k_exog_states, dtype=floatX))

        self.ssm["transition", :, :] = pt.linalg.block_diag(*transition_blocks)

        # Selection matrix (R)
        for i in range(self.k_factors):
            self.ssm["selection", i, i] = 1.0

        if self.error_order > 0:
            for i in range(self.k_endog):
                row = max(self.factor_order, 1) * self.k_factors + i
                col = self.k_factors + i
                self.ssm["selection", row, col] = 1.0

        if self.exog_flag and self.exog_innovations:
            row_start = self.k_states - self.k_exog_states
            row_end = self.k_states

            if self.error_order > 0:
                col_start = self.k_factors + self.k_endog
                col_end = self.k_factors + self.k_endog + self.k_exog_states
            else:
                col_start = self.k_factors
                col_end = self.k_factors + self.k_exog_states

            self.ssm["selection", row_start:row_end, col_start:col_end] = pt.eye(
                self.k_exog_states, dtype=floatX
            )

        factor_cov = pt.eye(self.k_factors, dtype=floatX)

        # Handle error_sigma and error_cov depending on error_cov_type
        if self.error_cov_type == "scalar":
            error_sigma = self.make_and_register_variable("error_sigma", shape=(), dtype=floatX)
            error_cov = pt.eye(self.k_endog) * error_sigma
        elif self.error_cov_type == "diagonal":
            error_sigma = self.make_and_register_variable(
                "error_sigma", shape=(self.k_endog,), dtype=floatX
            )
            error_cov = pt.diag(error_sigma)
        elif self.error_cov_type == "unstructured":
            error_cov = self.make_and_register_variable(
                "error_cov", shape=(self.k_endog, self.k_endog), dtype=floatX
            )

        # State covariance matrix (Q)
        if self.error_order > 0:
            if self.exog_flag and self.exog_innovations:
                # Include AR noise in state vector
                beta_sigma = self.make_and_register_variable(
                    "beta_sigma", shape=(self.k_exog_states,), dtype=floatX
                )
                exog_cov = pt.diag(beta_sigma)
                self.ssm["state_cov", :, :] = pt.linalg.block_diag(factor_cov, error_cov, exog_cov)
            elif self.exog_flag and not self.exog_innovations:
                exog_cov = pt.zeros((self.k_exog_states, self.k_exog_states), dtype=floatX)
                self.ssm["state_cov", :, :] = pt.linalg.block_diag(factor_cov, error_cov, exog_cov)
            elif not self.exog_flag:
                self.ssm["state_cov", :, :] = pt.linalg.block_diag(factor_cov, error_cov)
        else:
            if self.exog_flag and self.exog_innovations:
                beta_sigma = self.make_and_register_variable(
                    "beta_sigma", shape=(self.k_exog_states,), dtype=floatX
                )
                exog_cov = pt.diag(beta_sigma)
                self.ssm["state_cov", :, :] = pt.linalg.block_diag(factor_cov, exog_cov)
            elif self.exog_flag and not self.exog_innovations:
                exog_cov = pt.zeros((self.k_exog_states, self.k_exog_states), dtype=floatX)
                self.ssm["state_cov", :, :] = pt.linalg.block_diag(factor_cov, exog_cov)
            elif not self.exog_flag:
                # Only latent factor in the state
                self.ssm["state_cov", :, :] = factor_cov

        # Observation covariance matrix (H)
        if self.error_order > 0:
            if self.measurement_error:
                sigma_obs = self.make_and_register_variable(
                    "sigma_obs", shape=(self.k_endog,), dtype=floatX
                )
                self.ssm["obs_cov", :, :] = pt.diag(sigma_obs)
            # else: obs_cov remains zero (no measurement noise and idiosyncratic noise captured in state)
        else:
            if self.measurement_error:
                # TODO: check this decision
                # in this case error_order = 0, so there is no error term in the state, so the sigma error could not be added there
                # Idiosyncratic + measurement error
                sigma_obs = self.make_and_register_variable(
                    "sigma_obs", shape=(self.k_endog,), dtype=floatX
                )
                total_obs_var = error_sigma**2 + sigma_obs**2
                self.ssm["obs_cov", :, :] = pt.diag(pt.sqrt(total_obs_var))
            else:
                self.ssm["obs_cov", :, :] = pt.diag(error_sigma)



import pytensor

ALL_STATE_DIM = "state"
ALL_STATE_AUX_DIM = "state_aux"
OBS_STATE_DIM = "observed_state"
OBS_STATE_AUX_DIM = "observed_state_aux"
SHOCK_DIM = "shock"
SHOCK_AUX_DIM = "shock_aux"
TIME_DIM = "time"
AR_PARAM_DIM = "lag_ar"
MA_PARAM_DIM = "lag_ma"
SEASONAL_AR_PARAM_DIM = "seasonal_lag_ar"
SEASONAL_MA_PARAM_DIM = "seasonal_lag_ma"
ETS_SEASONAL_DIM = "seasonal_lag"
FACTOR_DIM = "factor"
ERROR_AR_PARAM_DIM = "error_lag_ar"
EXOG_STATE_DIM = "exogenous"

NEVER_TIME_VARYING = ["initial_state", "initial_state_cov", "a0", "P0"]
VECTOR_VALUED = ["initial_state", "state_intercept", "obs_intercept", "a0", "c", "d"]

MISSING_FILL = -9999.0
JITTER_DEFAULT = 1e-8 if pytensor.config.floatX.endswith("64") else 1e-6

FILTER_OUTPUT_TYPES = ["filtered", "predicted", "smoothed"]

MATRIX_NAMES = ["x0", "P0", "c", "d", "T", "Z", "R", "H", "Q"]
LONG_MATRIX_NAMES = [
    "initial_state",
    "initial_state_cov",
    "state_intercept",
    "obs_intercept",
    "transition",
    "design",
    "selection",
    "obs_cov",
    "state_cov",
]

SHORT_NAME_TO_LONG = dict(zip(MATRIX_NAMES, LONG_MATRIX_NAMES))
LONG_NAME_TO_SHORT = dict(zip(LONG_MATRIX_NAMES, MATRIX_NAMES))

FILTER_OUTPUT_NAMES = [
    "filtered_states",
    "predicted_states",
    "filtered_covariances",
    "predicted_covariances",
    "predicted_observed_states",
    "predicted_observed_covariances",
]

SMOOTHER_OUTPUT_NAMES = ["smoothed_states", "smoothed_covariances"]
OBSERVED_OUTPUT_NAMES = ["predicted_observed_states", "predicted_observed_covariances"]

MATRIX_DIMS = {
    "x0": (ALL_STATE_DIM,),
    "P0": (ALL_STATE_DIM, ALL_STATE_AUX_DIM),
    "c": (ALL_STATE_DIM,),
    "d": (OBS_STATE_DIM,),
    "T": (ALL_STATE_DIM, ALL_STATE_AUX_DIM),
    "Z": (OBS_STATE_DIM, ALL_STATE_DIM),
    "R": (ALL_STATE_DIM, SHOCK_DIM),
    "H": (OBS_STATE_DIM, OBS_STATE_AUX_DIM),
    "Q": (SHOCK_DIM, SHOCK_AUX_DIM),
}

FILTER_OUTPUT_DIMS = {
    "filtered_states": (TIME_DIM, ALL_STATE_DIM),
    "smoothed_states": (TIME_DIM, ALL_STATE_DIM),
    "predicted_states": (TIME_DIM, ALL_STATE_DIM),
    "filtered_covariances": (TIME_DIM, ALL_STATE_DIM, ALL_STATE_AUX_DIM),
    "smoothed_covariances": (TIME_DIM, ALL_STATE_DIM, ALL_STATE_AUX_DIM),
    "predicted_covariances": (TIME_DIM, ALL_STATE_DIM, ALL_STATE_AUX_DIM),
    "predicted_observed_states": (TIME_DIM, OBS_STATE_DIM),
    "predicted_observed_covariances": (TIME_DIM, OBS_STATE_DIM, OBS_STATE_AUX_DIM),
}

POSITION_DERIVATIVE_NAMES = ["level", "trend", "acceleration", "jerk", "snap", "crackle", "pop"]
SARIMAX_STATE_STRUCTURES = ["fast", "interpretable"]
